"""
This files tests the vent handlers not covered by other tests.
"""

from __future__ import annotations

import operator
from enum import Enum
from typing import TYPE_CHECKING, Callable

import numpy as np
import pytest
import qtpy.QtCore as QC
import qtpy.QtWidgets as QW
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.tests.unit.utils import (
    create_window,
    drag_mouse,
    scroll_wheel,
)
from plotpy.tools.selection import SelectTool

if TYPE_CHECKING:
    from plotpy.plot.plotwidget import PlotDialog, PlotWindow


class ZoomType(Enum):
    NO_ZOOM = (operator.eq, operator.eq, operator.eq, operator.eq)
    ZOOM_IN = (operator.gt, operator.lt, operator.lt, operator.gt)
    ZOOM_OUT = (operator.lt, operator.gt, operator.gt, operator.lt)


class ZoomEvent(Enum):
    ZOOM_WITH_MOUSE = 0
    ZOOM_WITH_WHEEL = 1


def result_zoom(zoom_type: ZoomType, event_type: ZoomEvent):
    """Wrapper used to parametrize the test_zoom function.

    Args:
        zoom_type: enum to specify the expected zoom.
        event_type: enum to specify the event type (drag or scroll).

    Returns:
        A function used to zoom in the plot and the expected zoom type.
    """
    if event_type is ZoomEvent.ZOOM_WITH_MOUSE:
        if zoom_type is ZoomType.NO_ZOOM:
            btn = QC.Qt.MouseButton.NoButton
            x_path = np.array([0.5, 0.8])
        elif zoom_type is ZoomType.ZOOM_IN:
            btn = QC.Qt.MouseButton.RightButton
            x_path = np.array([0.5, 0.8])
        else:
            btn = QC.Qt.MouseButton.RightButton
            x_path = np.array([0.5, 0.2])

        def _zoom_with_mouse(win: PlotWindow | PlotDialog):
            """Zoom in the plot by dragging the mouse while holding right click.

            Args:
                win: PlotWindow or PlotDialog instance.
            """
            drag_mouse(win, x_path, np.array([0.5, 0.5]), btn=btn)

        return _zoom_with_mouse, zoom_type

    if zoom_type is ZoomType.NO_ZOOM:
        angle_delta = 360
        mod = QC.Qt.KeyboardModifier.NoModifier
    elif zoom_type is ZoomType.ZOOM_IN:
        angle_delta = 360
        mod = QC.Qt.KeyboardModifier.ControlModifier
    else:
        angle_delta = -360
        mod = QC.Qt.KeyboardModifier.ControlModifier

    def _zoom_with_wheel(win: PlotWindow | PlotDialog):
        """Zoom in the plot by scrolling the mouse wheel while holding control.

        Args:
            win: PlotWindow or PlotDialog instance.
        """
        scroll_wheel(win, (0.5, 0.5), angle_delta, 0, mods=mod)

    return _zoom_with_wheel, zoom_type


@pytest.mark.parametrize(
    "zoom_func, zoom_type",
    [
        result_zoom(ZoomType.NO_ZOOM, ZoomEvent.ZOOM_WITH_MOUSE),
        result_zoom(ZoomType.NO_ZOOM, ZoomEvent.ZOOM_WITH_WHEEL),
        result_zoom(ZoomType.ZOOM_IN, ZoomEvent.ZOOM_WITH_MOUSE),
        result_zoom(ZoomType.ZOOM_IN, ZoomEvent.ZOOM_WITH_WHEEL),
        result_zoom(ZoomType.ZOOM_OUT, ZoomEvent.ZOOM_WITH_MOUSE),
        result_zoom(ZoomType.ZOOM_OUT, ZoomEvent.ZOOM_WITH_WHEEL),
    ],
)
def test_zoom(
    zoom_func: Callable[[QW.QApplication, PlotWindow | PlotDialog], None],
    zoom_type: ZoomType,
):
    """Test zooming on the plot by triggering several event types.

    Args:
        zoom_func: _description_
    """
    with qt_app_context(exec_loop=False):
        win, _ = create_window(SelectTool)
        win.show()

        plot = win.manager.get_plot()
        plot.unselect_all()

        axis_x, axis_y = plot.get_active_axes()

        x_min0, x_max0 = plot.get_axis_limits(axis_x)
        y_min0, y_max0 = plot.get_axis_limits(axis_y)

        zoom_func(win)

        x_min1, x_max1 = plot.get_axis_limits(axis_x)
        y_min1, y_max1 = plot.get_axis_limits(axis_y)

        c1, c2, c3, c4 = zoom_type.value
        assert c1(x_min1, x_min0)
        assert c2(x_max1, x_max0)
        assert c3(y_min1, y_min0)
        assert c4(y_max1, y_max0)

        exec_dialog(win)


def test_pan():
    """Test panning the plot by dragging the mouse while holding middle click."""
    with qt_app_context(exec_loop=False):
        win, _ = create_window(SelectTool)
        win.show()

        plot = win.manager.get_plot()
        plot.unselect_all()

        axis_x, axis_y = plot.get_active_axes()

        x_min0, x_max0 = plot.get_axis_limits(axis_x)
        y_min0, y_max0 = plot.get_axis_limits(axis_y)

        x_path, y_path = np.linspace(0.5, 0, 100), np.linspace(0.5, 0, 100)
        drag_mouse(win, x_path, y_path, btn=QC.Qt.MouseButton.MiddleButton)

        x_min1, x_max1 = plot.get_axis_limits(axis_x)
        y_min1, y_max1 = plot.get_axis_limits(axis_y)

        assert x_min1 > x_min0
        assert x_max1 > x_max0
        assert y_min1 > y_min0
        assert y_max1 > y_max0

        exec_dialog(win)


if __name__ == "__main__":
    test_zoom(*result_zoom(ZoomType.ZOOM_OUT, ZoomEvent.ZOOM_WITH_WHEEL))
    test_pan()



# -*- coding: utf-8 -*-# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Axes tools"""

from __future__ import annotations

from guidata.configtools import get_icon
from guidata.qthelpers import add_actions
from qtpy import QtWidgets as QW

from plotpy.config import _
from plotpy.items import Axes
from plotpy.tools.base import CommandTool
from plotpy.tools.shape import RectangularShapeTool


class AxisScaleTool(CommandTool):
    """
    A tool for changing the scale of plot axes.

    This tool provides options to switch between linear and logarithmic scales
    for both x and y axes.
    """

    def __init__(self, manager):
        """
        Initialize the AxisScaleTool.

        Args:
            manager: The plot manager.
        """
        super().__init__(
            manager, _("Scale"), icon=get_icon("log_log.png"), tip=None, toolbar_id=None
        )
        self.action.setEnabled(True)

    def create_action_menu(self, manager) -> QW.QMenu:
        """
        Create and return menu for the tool's action.

        Args:
            manager: The plot manager.

        Returns:
            A QMenu object containing scale options.
        """
        menu = QW.QMenu()
        group = QW.QActionGroup(manager.get_main())
        lin_lin = manager.create_action(
            "Lin Lin",
            icon=get_icon("lin_lin.png"),
            toggled=lambda state, x="lin", y="lin": self.set_scale(state, x, y),
        )
        lin_log = manager.create_action(
            "Lin Log",
            icon=get_icon("lin_log.png"),
            toggled=lambda state, x="lin", y="log": self.set_scale(state, x, y),
        )
        log_lin = manager.create_action(
            "Log Lin",
            icon=get_icon("log_lin.png"),
            toggled=lambda state, x="log", y="lin": self.set_scale(state, x, y),
        )
        log_log = manager.create_action(
            "Log Log",
            icon=get_icon("log_log.png"),
            toggled=lambda state, x="log", y="log": self.set_scale(state, x, y),
        )
        date_time = manager.create_action(
            "Date Time",
            toggled=lambda state, x="datetime", y="lin": self.set_scale(state, x, y),
        )
        self.scale_menu = {
            ("lin", "lin"): lin_lin,
            ("lin", "log"): lin_log,
            ("log", "lin"): log_lin,
            ("log", "log"): log_log,
            ("datetime", "lin"): date_time,
        }
        for obj in (group, menu):
            add_actions(obj, (lin_lin, lin_log, log_lin, log_log, date_time))
        return menu

    def update_status(self, plot) -> None:
        """
        Update the status of scale actions based on the current plot scales.

        Args:
            plot: The current plot object.
        """
        item = plot.get_active_item()
        active_scale = ("lin", "lin")
        if item is not None:
            xscale = plot.get_axis_scale(item.xAxis())
            yscale = plot.get_axis_scale(item.yAxis())
            active_scale = xscale, yscale
        for scale_type, scale_action in list(self.scale_menu.items()):
            if item is None:
                scale_action.setEnabled(False)
            else:
                scale_action.setEnabled(True)
                if active_scale == scale_type:
                    scale_action.setChecked(True)
                else:
                    scale_action.setChecked(False)

    def set_scale(self, checked: bool, xscale: str, yscale: str) -> None:
        """
        Set the scale of the active plot.

        Args:
            checked: Whether the action is checked.
            xscale: The scale for the x-axis ('lin', 'log', or 'datetime').
            yscale: The scale for the y-axis ('lin' or 'log').
        """
        if not checked:
            return
        plot = self.get_active_plot()
        if plot is not None:
            cur_xscale, cur_yscale = plot.get_scales()
            if cur_xscale != xscale or cur_yscale != yscale:
                plot.set_scales(xscale, yscale)


class PlaceAxesTool(RectangularShapeTool):
    """
    A tool for placing axes on the plot.

    This tool allows users to draw a rectangular shape to define
    the position and size of the axes on the plot.
    """

    TITLE: str = _("Axes")
    ICON: str = "gtaxes.png"
    SHAPE_STYLE_KEY: str = "shape/axes"

    def create_shape(self) -> tuple[Axes, int, int]:
        """
        Create an Axes shape.

        Returns:
            A tuple containing the Axes object and its handle indices.
        """
        shape = Axes((0, 1), (1, 1), (0, 0))
        self.set_shape_style(shape)
        return shape, 0, 2



# -*- coding: utf-8 -*-

from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import numpy as np
from guidata.dataset import update_dataset
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qwt import QwtPlotCurve

from plotpy.config import CONF, _
from plotpy.coords import canvas_to_axes
from plotpy.interfaces import (
    IBasePlotItem,
    ICurveItemType,
    ISerializableType,
    ITrackableItemType,
)
from plotpy.styles.base import SymbolParam
from plotpy.styles.curve import CurveParam

if TYPE_CHECKING:
    import guidata.io

    from plotpy.interfaces import IItemType
    from plotpy.plot import BasePlot
    from plotpy.styles.base import ItemParameters


SELECTED_SYMBOL_PARAM = SymbolParam()
SELECTED_SYMBOL_PARAM.read_config(CONF, "plot", "selected_curve_symbol")
SELECTED_SYMBOL = SELECTED_SYMBOL_PARAM.build_symbol()


def seg_dist(P: QC.QPointF, P0: QC.QPointF, P1: QC.QPointF) -> float:
    """Compute distance between point P and segment (P0, P1)

    Args:
        P: QPointF instance
        P0: QPointF instance
        P1: QPointF instance

    Returns:
        Distance between point P and segment (P0, P1)

    .. note::

        If P orthogonal projection on (P0, P1) is outside segment bounds, return
        either distance to P0 or to P1 (the closest one)
    """
    u = QC.QLineF(P0, P).length()
    if P0 == P1:
        return u
    else:
        angle = QC.QLineF(P0, P).angleTo(QC.QLineF(P0, P1)) * np.pi / 180
        projection = u * np.cos(angle)
        if projection > QC.QLineF(P0, P1).length():
            return QC.QLineF(P1, P).length()
        elif projection < 0:
            return QC.QLineF(P0, P).length()
        else:
            return abs(u * np.sin(angle))


def seg_dist_v(
    P: tuple[float, float], X0: float, Y0: float, X1: float, Y1: float
) -> tuple[int, float]:
    """Compute distance between point P and segment (X0, Y0), (X1, Y1)

    Args:
        P: Point
        X0: X coordinate of first point
        Y0: Y coordinate of first point
        X1: X coordinate of second point
        Y1: Y coordinate of second point

    Returns:
        tuple: Tuple with two elements: (index, distance)

    .. note::

        This is the vectorized version of ``seg_dist`` function
    """
    V = np.zeros((X0.shape[0], 2), float)
    PP = np.zeros((X0.shape[0], 2), float)
    PP[:, 0] = X0
    PP[:, 1] = Y0
    V[:, 0] = X1 - X0
    V[:, 1] = Y1 - Y0
    dP = np.array(P).reshape(1, 2) - PP
    norm2V = (V**2).sum(axis=1)
    nV = np.sqrt(norm2V).clip(1e-12)  # clip: avoid division by zero
    w2 = V / nV[:, np.newaxis]
    w = np.array([-w2[:, 1], w2[:, 0]]).T
    distances = np.fabs((dP * w).sum(axis=1))
    ix = distances.argmin()
    return ix, distances[ix]


class CurveItem(QwtPlotCurve):
    """Curve item

    Args:
        curveparam: Curve parameters
    """

    __implements__ = (IBasePlotItem, ISerializableType)

    _readonly = False
    _private = False
    _icon_name = "curve.png"

    def __init__(self, curveparam: CurveParam | None = None) -> None:
        super().__init__()
        if curveparam is None:
            self.param = CurveParam(_("Curve"), icon="curve.png")
        else:
            self.param = curveparam
        self.selected = False
        self.immutable = True  # set to false to allow moving points around
        self._x = None
        self._y = None
        self.update_params()

    def _get_visible_axis_min(self, axis_id: int, axis_data: np.ndarray) -> float:
        """Return axis minimum excluding zero and negative values when
        corresponding plot axis scale is logarithmic

        Args:
            axis_id: Axis ID
            axis_data: Axis data

        Returns:
            Axis minimum
        """
        if self.plot().get_axis_scale(axis_id) == "log":
            if len(axis_data[axis_data > 0]) == 0:
                return 0.0
            else:
                return axis_data[axis_data > 0].min()
        else:
            return axis_data.min()

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the data

        Returns:
            Bounding rectangle of the data
        """
        plot = self.plot()
        if plot is not None and "log" in (
            plot.get_axis_scale(self.xAxis()),
            plot.get_axis_scale(self.yAxis()),
        ):
            x, y = self._x, self._y
            xf, yf = x[np.isfinite(x)], y[np.isfinite(y)]
            xmin = self._get_visible_axis_min(self.xAxis(), xf)
            ymin = self._get_visible_axis_min(self.yAxis(), yf)
            return QC.QRectF(xmin, ymin, xf.max() - xmin, yf.max() - ymin)
        else:
            return QwtPlotCurve.boundingRect(self)

    def types(self) -> tuple[type[IItemType], ...]:
        """Returns a group or category for this item.
        This should be a tuple of class objects inheriting from IItemType

        Returns:
            tuple: Tuple of class objects inheriting from IItemType
        """
        return (ICurveItemType, ITrackableItemType, ISerializableType)

    def get_icon_name(self) -> str:
        """Return the icon name

        Returns:
            Icon name
        """
        return self._icon_name

    def set_icon_name(self, icon_name: str) -> None:
        """Set the icon name

        Args:
            icon_name: Icon name
        """
        self._icon_name = icon_name

    def set_selectable(self, state: bool) -> None:
        """Set item selectable state

        Args:
            state: True if item is selectable, False otherwise
        """
        self._can_select = state

    def set_resizable(self, state: bool) -> None:
        """Set item resizable state
        (or any action triggered when moving an handle, e.g. rotation)

        Args:
            state: True if item is resizable, False otherwise
        """
        self._can_resize = state

    def set_movable(self, state: bool) -> None:
        """Set item movable state

        Args:
            state: True if item is movable, False otherwise
        """
        self._can_move = state

    def set_rotatable(self, state: bool) -> None:
        """Set item rotatable state

        Args:
            state: True if item is rotatable, False otherwise
        """
        self._can_rotate = state

    def can_select(self) -> bool:
        """
        Returns True if this item can be selected

        Returns:
            bool: True if item can be selected, False otherwise
        """
        return True

    def can_resize(self) -> bool:
        """
        Returns True if this item can be resized

        Returns:
            bool: True if item can be resized, False otherwise
        """
        return False

    def can_rotate(self) -> bool:
        """
        Returns True if this item can be rotated

        Returns:
            bool: True if item can be rotated, False otherwise
        """
        return False

    def can_move(self) -> bool:
        """
        Returns True if this item can be moved

        Returns:
            bool: True if item can be moved, False otherwise
        """
        return False

    def __reduce__(self) -> tuple[type, tuple, tuple]:
        """Return state information for pickling"""
        state = (self.param, self._x, self._y, self.z())
        res = (CurveItem, (), state)
        return res

    def __setstate__(self, state: tuple) -> None:
        """Restore state information for unpickling"""
        param, x, y, z = state
        self.param = param
        self.set_data(x, y)
        self.setZ(z)
        self.update_params()

    def serialize(
        self,
        writer: guidata.io.HDF5Writer | guidata.io.INIWriter | guidata.io.JSONWriter,
    ) -> None:
        """Serialize object to HDF5 writer

        Args:
            writer: HDF5, INI or JSON writer
        """
        writer.write(self._x, group_name="Xdata")
        writer.write(self._y, group_name="Ydata")
        writer.write(self.z(), group_name="z")
        self.param.update_param(self)
        writer.write(self.param, group_name="curveparam")

    def deserialize(
        self,
        reader: guidata.io.HDF5Reader | guidata.io.INIReader | guidata.io.JSONReader,
    ) -> None:
        """Deserialize object from HDF5 reader

        Args:
            reader: HDF5, INI or JSON reader
        """
        self.param = CurveParam(_("Curve"), icon="curve.png")
        reader.read("curveparam", instance=self.param)
        x = reader.read(group_name="Xdata", func=reader.read_array)
        y = reader.read(group_name="Ydata", func=reader.read_array)
        self.set_data(x, y)
        self.setZ(reader.read("z"))
        self.update_params()

    def set_readonly(self, state: bool) -> None:
        """Set object readonly state

        Args:
            state: True if object is readonly, False otherwise
        """
        self._readonly = state

    def is_readonly(self) -> bool:
        """Return object readonly state

        Returns:
            bool: True if object is readonly, False otherwise
        """
        return self._readonly

    def set_private(self, state: bool) -> None:
        """Set object as private

        Args:
            state: True if object is private, False otherwise
        """
        self._private = state

    def is_private(self) -> bool:
        """Return True if object is private

        Returns:
            bool: True if object is private, False otherwise
        """
        return self._private

    def invalidate_plot(self) -> None:
        """Invalidate the plot to force a redraw"""
        plot = self.plot()
        if plot is not None:
            plot.invalidate()

    def select(self) -> None:
        """Select the object and eventually change its appearance to highlight the
        fact that it's selected
        """
        self.selected = True
        plot = self.plot()
        if plot is not None:
            plot.blockSignals(True)
        self.setSymbol(SELECTED_SYMBOL)
        if plot is not None:
            plot.blockSignals(False)
        self.invalidate_plot()

    def unselect(self) -> None:
        """Unselect the object and eventually restore its original appearance to
        highlight the fact that it's not selected anymore
        """
        self.selected = False
        # Restoring initial curve parameters:
        self.param.update_item(self)
        self.invalidate_plot()

    def get_data(self) -> tuple[np.ndarray, np.ndarray]:
        """Return curve data x, y (NumPy arrays)

        Returns:
            tuple: Tuple with two elements: x and y NumPy arrays
        """
        assert isinstance(self._x, np.ndarray) and isinstance(self._y, np.ndarray)
        return self._x, self._y

    def update_data(self) -> None:
        """Update curve data with current arrays."""
        if isinstance(self._x, np.ndarray) and isinstance(self._y, np.ndarray):
            self._setData(self._x, self._y)

    def dsamp(self, data: np.ndarray) -> np.ndarray:
        """Downsample data

        Args:
            data: Data to downsample

        Returns:
            Downsampled data
        """
        if self.param.use_dsamp and self.param.dsamp_factor > 1:
            return data[:: self.param.dsamp_factor]
        return data

    def _setData(self, x: np.ndarray, y: np.ndarray) -> None:
        """Wrapper around QwtPlotCurve.setData() to handle downsampling"""
        return super().setData(self.dsamp(x), self.dsamp(y))

    def set_data(self, x: np.ndarray, y: np.ndarray) -> None:
        """Set curve data

        Args:
            x: X data
            y: Y data
            decimated_data: Set to True if CurveItem X and Y arrays are already set and
             this method is called to update decimated data (i.e. only update 1/N value
             with N set in CurveItem.param.decimation).
        """
        self._x = np.asarray(x)
        self._y = np.asarray(y)
        self._setData(self._x, self._y)

    def is_empty(self) -> bool:
        """Return True if the item is empty

        Returns:
            True if the item is empty, False otherwise
        """
        return self._x is None or self._y is None or self._y.size == 0

    def hit_test(self, pos: QC.QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        if self.is_empty():
            return sys.maxsize, 0, False, None
        plot = self.plot()
        ax = self.xAxis()
        ay = self.yAxis()
        px = plot.invTransform(ax, pos.x())
        py = plot.invTransform(ay, pos.y())
        # On cherche les 4 points qui sont les plus proches en X et en Y
        # avant et après ie tels que p1x < x < p2x et p3y < y < p4y
        tmpx = self._x - px
        tmpy = self._y - py
        if np.count_nonzero(tmpx) != len(tmpx) or np.count_nonzero(tmpy) != len(tmpy):
            # Avoid dividing by zero warning when computing dx or dy
            return sys.maxsize, 0, False, None
        dx = 1 / tmpx
        dy = 1 / tmpy
        i0 = dx.argmin()
        i1 = dx.argmax()
        i2 = dy.argmin()
        i3 = dy.argmax()
        t = np.array((i0, i1, i2, i3))
        t2 = (t + 1).clip(0, self._x.shape[0] - 1)
        i, _d = seg_dist_v((px, py), self._x[t], self._y[t], self._x[t2], self._y[t2])
        i = t[i]
        # Recalcule la distance dans le répère du widget
        p0x = plot.transform(ax, self._x[i])
        p0y = plot.transform(ay, self._y[i])
        if i + 1 >= self._x.shape[0]:
            p1x = p0x
            p1y = p0y
        else:
            p1x = plot.transform(ax, self._x[i + 1])
            p1y = plot.transform(ay, self._y[i + 1])
        distance = seg_dist(QC.QPointF(pos), QC.QPointF(p0x, p0y), QC.QPointF(p1x, p1y))
        final_index = i // (int(not self.param.use_dsamp) or self.param.dsamp_factor)
        return distance, final_index, False, None

    def get_closest_coordinates(self, x: float, y: float) -> tuple[float, float]:
        """
        Get the closest coordinates to the given point

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            tuple[float, float]: Closest coordinates
        """
        plot = self.plot()
        ax = self.xAxis()
        ay = self.yAxis()
        xc = plot.transform(ax, x)
        yc = plot.transform(ay, y)
        _distance, i, _inside, _other = self.hit_test(QC.QPointF(xc, yc))
        point = self.sample(i)
        return point.x(), point.y()

    def get_coordinates_label(self, x: float, y: float) -> str:
        """
        Get the coordinates label for the given coordinates

        Args:
            x: X coordinate
            y: Y coordinate

        Returns:
            str: Coordinates label
        """
        title = self.title().text()

        # Use the plot's consolidated coordinate formatting method
        plot: BasePlot = self.plot()
        if plot is not None:
            # Format using the plot's consolidated coordinate formatting method
            return plot.format_coordinate_values(
                x, y, self.xAxis(), self.yAxis(), title
            )
        else:
            return f"{title}:<br>x = {x:g}<br>y = {y:g}"

    def get_closest_x(self, xc: float) -> tuple[float, float]:
        """
        Get the closest point to the given x coordinate

        Args:
            xc: X coordinate

        Returns:
            tuple[float, float]: Closest point coordinates
        """
        # We assume X is sorted, otherwise we'd need :
        # argmin(abs(x-xc))
        i = self._x.searchsorted(xc)
        if i > 0 and np.fabs(self._x[i - 1] - xc) < np.fabs(self._x[i] - xc):
            return self._x[i - 1], self._y[i - 1]
        return self._x[i], self._y[i]

    def move_local_point_to(
        self, handle: int, pos: QC.QPointF, ctrl: bool = None
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        if self.immutable:
            return
        if handle < 0 or handle > self._x.shape[0]:
            return
        x, y = canvas_to_axes(self, pos)
        self._x[handle] = x
        self._y[handle] = y
        self._setData(self._x, self._y)
        self.plot().replot()

    def move_local_shape(self, old_pos: QC.QPointF, new_pos: QC.QPointF) -> None:
        """Translate the shape such that old_pos becomes new_pos in canvas coordinates

        Args:
            old_pos: Old position
            new_pos: New position
        """
        nx, ny = canvas_to_axes(self, new_pos)
        ox, oy = canvas_to_axes(self, old_pos)
        self._x += nx - ox
        self._y += ny - oy
        self._setData(self._x, self._y)

    def move_with_selection(self, delta_x: float, delta_y: float) -> None:
        """Translate the item together with other selected items

        Args:
            delta_x: Translation in plot coordinates along x-axis
            delta_y: Translation in plot coordinates along y-axis
        """
        self._x += delta_x
        self._y += delta_y
        self._setData(self._x, self._y)

    def update_params(self):
        """Update item parameters (object properties) from datasets"""
        self.param.update_item(self)
        if self.selected:
            self.select()

    def update_item_parameters(self) -> None:
        """Update item parameters (dataset) from object properties"""
        self.param.update_param(self)

    def get_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Appends datasets to the list of DataSets describing the parameters
        used to customize apearance of this item

        Args:
            itemparams: Item parameters
        """
        itemparams.add("CurveParam", self, self.param)

    def set_item_parameters(self, itemparams: ItemParameters) -> None:
        """
        Change the appearance of this item according
        to the parameter set provided

        Args:
            itemparams: Item parameters
        """
        update_dataset(self.param, itemparams.get("CurveParam"), visible_only=True)
        self.update_params()


assert_interfaces_valid(CurveItem)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Z-axis log scale tool test"""

# guitest: show

from __future__ import annotations

import numpy as np
from guidata.qthelpers import qt_app_context, qt_wait

from plotpy.builder import make
from plotpy.plot import PlotDialog, PlotOptions
from plotpy.tests.data import gen_2d_gaussian
from plotpy.tools import ZAxisLogTool


class MyPlotDialog(PlotDialog):
    def __init__(self) -> None:
        """Reimplement PlotDialog method"""
        super().__init__(
            title="Z-axis log scale tool test",
            toolbar=True,
            options=PlotOptions(type="image"),
        )
        # No need to add the tools to the manager, they are automatically added
        # when the `register_curve_tools` or `register_image_tools` method is called
        self.setup_items()

    def setup_items(self) -> None:
        """Setup items"""
        plot = self.get_plot()
        data = gen_2d_gaussian(512, np.uint16)
        item = make.image(data, title="Image", colormap="plasma")
        plot.add_item(item)
        plot.set_active_item(item)
        item.unselect()


def test_zaxislogtool() -> None:
    """Test the Z-axis log scale tool"""
    with qt_app_context(exec_loop=True):
        win = MyPlotDialog()
        win.show()
        tool = win.manager.get_tool(ZAxisLogTool)
        qt_wait(1, except_unattended=True)
        for _index in range(2):
            tool.activate()
            qt_wait(1, except_unattended=True)


if __name__ == "__main__":
    test_zaxislogtool()



# -*- coding: utf-8

"""
Select dialog
-------------

The `selectdialog` module provides a dialog box to select an area of the plot
using a tool:

    * :py:class:`.widgets.selectdialog.SelectDialog`: dialog box
    * :py:func:`.widgets.selectdialog.select_with_shape_tool`: function to
     select an area with a shape tool and return the rectangle
    * :py:func:`.widgets.selectdialog.set_items_unselectable`: function to set
     items unselectable except for the given item

Example: get segment
~~~~~~~~~~~~~~~~~~~~

.. literalinclude:: ../../plotpy/tests/tools/test_get_segment.py

Example: get rectangle
~~~~~~~~~~~~~~~~~~~~~~

.. literalinclude:: ../../plotpy/tests/tools/test_get_rectangle.py

Example: get rectangle with SVG
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

.. literalinclude:: ../../plotpy/tests/tools/test_get_rectangle_with_svg.py

Reference
~~~~~~~~~

.. autoclass:: SelectDialog
   :members:
.. autofunction:: select_with_shape_tool
.. autofunction:: set_items_unselectable
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from guidata.configtools import get_icon
from guidata.qthelpers import exec_dialog
from qtpy import QtWidgets as QW
from qtpy.QtWidgets import QWidget  # only to help intersphinx find QWidget
from qwt.plot import QwtPlotItem

from plotpy.items import AbstractShape, ImageItem
from plotpy.panels.base import PanelWidget
from plotpy.plot import BasePlot, PlotDialog, PlotOptions
from plotpy.tools import RectangularShapeTool, SelectTool

if TYPE_CHECKING:
    from plotpy.panels.base import PanelWidget


class SelectDialog(PlotDialog):
    """Plot dialog box to select an area of the plot using a tool

    Args:
        parent: parent widget
        toolbar: show/hide toolbar
        options: plot options
        panels: additionnal panels
        auto_tools: If True, the plot tools are automatically registered.
         If False, the user must register the tools manually.
        title: The window title
        icon: The window icon
        edit: If True, the plot is editable
    """

    def __init__(
        self,
        parent: QWidget | None = None,
        toolbar: bool = False,
        options: PlotOptions | dict[str, Any] | None = None,
        panels: list[PanelWidget] | None = None,
        auto_tools: bool = True,
        title: str = "PlotPy",
        icon: str = "plotpy.svg",
        edit: bool = False,
    ) -> None:
        super().__init__(
            parent, toolbar, options, panels, auto_tools, title, icon, edit
        )
        self.sel_tool: RectangularShapeTool | None = None

    def set_image_and_tool(
        self, item: ImageItem, toolclass: RectangularShapeTool, **kwargs
    ) -> None:
        """Set the image item to be displayed and the tool to be used

        Args:
            item: Image item
            toolclass: Tool class
            kwargs: Keyword arguments for the tool class
        """
        default = self.manager.add_tool(SelectTool)
        self.manager.set_default_tool(default)
        self.sel_tool: RectangularShapeTool = self.manager.add_tool(
            toolclass,
            switch_to_default_tool=True,
            **kwargs,
        )  # pylint: disable=attribute-defined-outside-init
        self.sel_tool.activate()
        set_ok_btn_enabled = self.button_box.button(QW.QDialogButtonBox.Ok).setEnabled
        set_ok_btn_enabled(False)
        self.sel_tool.SIG_TOOL_JOB_FINISHED.connect(lambda: set_ok_btn_enabled(True))
        plot = self.get_plot()
        plot.add_item(item)
        plot.set_active_item(item)
        item.set_selectable(False)
        item.set_readonly(True)
        plot.unselect_item(item)

    def get_new_shape(self) -> AbstractShape:
        """Get newly created shape

        Returns:
            Newly created shape
        """
        return self.sel_tool.get_last_final_shape()


def set_items_unselectable(plot: BasePlot, except_item: QwtPlotItem = None) -> None:
    """Set items unselectable except for the given item"""
    for item_i in plot.get_items():
        if except_item is None:
            item_i.set_selectable(False)
        else:
            item_i.set_selectable(item_i is except_item)


def select_with_shape_tool(
    parent: QW.QWidget,
    toolclass: RectangularShapeTool,
    item: ImageItem,
    title: str = None,
    size: tuple[int, int] = None,
    other_items: list[QwtPlotItem] = [],
    tooldialogclass: SelectDialog = SelectDialog,
    toolbar: bool = False,
    options: PlotOptions | dict[str, Any] | None = None,
    icon=None,
    **kwargs,
) -> AbstractShape:
    """Select an area with a shape tool and return the rectangle

    Args:
        parent: Parent widget
        toolclass: Tool class
        item: Image item
        title: Dialog title
        size: Dialog size
        other_items: Other items to be displayed
        tooldialogclass: Tool dialog class
        toolbar: show/hide toolbar
        options: plot options
        icon: Icon
        kwargs: Keyword arguments for the tool class

    Returns:
        Selected shape
    """
    if title is None:
        title = "Select an area then press OK to accept"
    if icon is not None:
        icon = get_icon(icon) if isinstance(icon, str) else icon
    win: SelectDialog = tooldialogclass(
        parent, title=title, edit=True, toolbar=toolbar, options=options, icon=icon
    )
    win.set_image_and_tool(item, toolclass, **kwargs)
    plot = win.get_plot()
    for other_item in other_items:
        plot.add_item(other_item)
    set_items_unselectable(plot)
    if size is not None:
        win.resize(*size)
    win.show()
    if exec_dialog(win):
        return win.get_new_shape()
    return None



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
plotplugin
==========

A plotpy plot widget plugin for Qt Designer
"""

from plotpy.widgets.qtdesigner import create_qtdesigner_plugin

Plugin = create_qtdesigner_plugin(
    "plotpy", "plotpy.plot", "PlotWidget", icon="curve.png"
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
base
----

The `base` module provides base objects for internal use of the
:mod:`.widgets` package.

"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Union

from guidata.configtools import get_icon
from guidata.qthelpers import create_toolbutton
from qtpy import QtWidgets as QW

from plotpy._scaler import INTERP_LINEAR
from plotpy.config import _
from plotpy.items import TrImageItem
from plotpy.plot import PlotOptions, PlotWidget

if TYPE_CHECKING:
    import numpy as np

    from plotpy.plot.base import BasePlot
    from plotpy.plot.manager import PlotManager
    from plotpy.widgets.fliprotate import FlipRotateWidget
    from plotpy.widgets.rotatecrop import RotateCropWidget

    TransfWidget = Union[FlipRotateWidget, RotateCropWidget]


class BaseTransform:
    """Base transform class (for manipulating TrImageItem objects)

    This is to be mixed with a class providing the get_plot method,
    like PlotDialog, or BaseTransformWidget (see below)

    Args:
        parent (QWidget): Parent widget
        manager (PlotManager): Plot manager
    """

    def __init__(self, parent: QW.QWidget, manager: PlotManager) -> None:
        self.parent = parent
        self.manager = manager
        self.item: TrImageItem = None
        self.item_original_state: tuple[bool, bool, bool, bool] = None
        self.item_original_crop: tuple[float, float, float, float] = None
        self.item_original_transform: tuple[float, float, float] = None
        self.output_array: np.ndarray = None

    # ------Public API----------------------------------------------------------
    def get_result(self) -> np.ndarray:
        """Return the result array

        Returns:
            numpy.ndarray: Result array
        """
        return self.output_array

    def set_item(self, item: TrImageItem) -> None:
        """Set associated item

        Args:
            item (TrImageItem): Image item
        """
        assert isinstance(item, TrImageItem)
        self.item = item
        self.item_original_state = (
            item.can_select(),
            item.can_move(),
            item.can_resize(),
            item.can_rotate(),
        )
        self.item_original_crop = item.get_crop()
        self.item_original_transform = item.get_transform()

        self.item.set_selectable(True)
        self.item.set_movable(True)
        self.item.set_resizable(False)
        self.item.set_rotatable(True)

        item.set_lut_threshold(2.0)
        item.set_interpolation(INTERP_LINEAR)
        plot = self.manager.get_plot()
        plot.add_item(self.item)

        # Setting the item as active item (even if the cropping rectangle item
        # will also be set as active item just below), for the image tools to
        # register this item (contrast, ...):
        plot.set_active_item(self.item)
        self.item.unselect()

    def unset_item(self) -> None:
        """Unset the associated item, freeing memory"""
        plot = self.manager.get_plot()
        plot.del_item(self.item)
        self.item = None

    def reset(self) -> None:
        """Reset crop/transform image settings"""
        self.item.set_crop(*self.item_original_crop)
        self.item.set_transform(*self.item_original_transform)
        self.reset_transformation()
        self.apply_transformation()

    def reset_transformation(self) -> None:
        """Reset transformation"""
        raise NotImplementedError

    def apply_transformation(self) -> None:
        """Apply transformation, e.g. crop or rotate"""
        raise NotImplementedError

    def compute_transformation(self) -> np.ndarray:
        """Compute transformation, return compute output array

        Returns:
            numpy.ndarray: Output array
        """
        raise NotImplementedError

    # ------Private API---------------------------------------------------------
    def restore_original_state(self) -> None:
        """Restore item original state"""
        select, move, resize, rotate = self.item_original_state
        self.item.set_selectable(select)
        self.item.set_movable(move)
        self.item.set_resizable(resize)
        self.item.set_rotatable(rotate)

    def accept_changes(self) -> None:
        """Computed rotated/cropped array and apply changes to item"""
        self.restore_original_state()
        self.apply_transformation()
        self.output_array = self.compute_transformation()
        # Ignoring image position changes
        pos_x0, pos_y0, _angle, sx, sy, hf, vf = self.item_original_transform
        _pos_x0, _pos_y0, angle, _sx, _sy, hf, vf = self.item.get_transform()
        self.item.set_transform(pos_x0, pos_y0, angle, sx, sy, hf, vf)

    def reject_changes(self) -> None:
        """Restore item original transform settings"""
        self.restore_original_state()
        self.item.set_crop(*self.item_original_crop)
        self.item.set_transform(*self.item_original_transform)


class BaseTransformWidget(QW.QWidget):
    """Base transform widget: see for example rotatecrop.py

    Args:
        parent (QWidget): Parent widget
        toolbar (bool | None): Whether to show toolbar. Defaults to False.
        options (dict | None): Plot options. Defaults to None.
    """

    def __init__(
        self,
        parent: QW.QWidget,
        toolbar: bool = False,
        options: PlotOptions | dict[str, Any] | None = None,
    ) -> None:
        super().__init__(parent)
        if isinstance(options, dict):
            options = PlotOptions(**options)
        options = options if options is not None else PlotOptions()
        options.type = "image"
        self.plot_widget = PlotWidget(self, options=options, toolbar=toolbar)
        self.plot_widget.manager.register_all_image_tools()
        hlayout = QW.QHBoxLayout()
        self.add_buttons_to_layout(hlayout)

        vlayout = QW.QVBoxLayout()
        vlayout.addWidget(self.plot_widget)
        vlayout.addLayout(hlayout)
        self.setLayout(vlayout)

    def get_plot(self) -> BasePlot:
        """Required for BaseTransformMixin

        Returns:
            BasePlot: Plot widget
        """
        return self.plot_widget.get_plot()

    def add_buttons_to_layout(
        self, layout: QW.QBoxLayout, apply=True, reset=True
    ) -> None:
        """Add tool buttons to layout

        Args:
            layout (QBoxLayout): Layout
            apply (bool | None): Add apply button. Defaults to True.
            reset (bool | None): Add reset button. Defaults to True.
        """
        if reset:
            self.__add_reset_button(layout)
        if apply:
            self.__add_apply_button(layout)

    def __add_apply_button(self, layout: QW.QBoxLayout) -> None:
        """Add the standard apply button

        Args:
            layout (QBoxLayout): Layout
        """
        apply_btn = create_toolbutton(
            self.plot_widget,
            text=_("Apply"),
            icon=get_icon("apply.png"),
            triggered=self.apply_transformation,
            autoraise=False,
        )
        layout.addWidget(apply_btn)
        layout.addStretch()

    def __add_reset_button(self, layout: QW.QBoxLayout) -> None:
        """Add the standard reset button

        Args:
            layout (QBoxLayout): Layout
        """
        edit_options_btn = create_toolbutton(
            self.plot_widget,
            text=_("Reset"),
            icon=get_icon("eraser.png"),
            triggered=self.reset,
            autoraise=False,
        )
        layout.addWidget(edit_options_btn)
        layout.addStretch()

    def reset(self) -> None:
        """Reset crop/transform image settings"""
        return NotImplementedError

    def apply_transformation(self) -> None:
        """Apply transformation, e.g. crop or rotate"""
        return NotImplementedError


class BaseMultipleTransformWidget(QW.QTabWidget):
    """Base Multiple Transform Widget

    Transform several :py:class:`.image.TrImageItem` plot items

    Args:
        parent: Parent widget
        options: Plot options. Defaults to None.
    """

    TRANSFORM_WIDGET_CLASS: TransfWidget  # to be defined in subclass

    def __init__(self, parent: QW.QWidget, options: PlotOptions | None = None):
        QW.QTabWidget.__init__(self, parent)
        self.options = options
        self.output_arrays = None

    def set_items(self, *items: TrImageItem) -> None:
        """Set the associated items

        Args:
            *items (TrImageItem): Items
        """
        for item in items:
            self.add_item(item)

    def add_item(self, item: TrImageItem) -> TransfWidget:
        """Add item to widget

        Args:
            item (TrImageItem): Item
        """
        widget: TransfWidget = self.TRANSFORM_WIDGET_CLASS(self, options=self.options)
        widget.transform.set_item(item)
        self.addTab(widget, item.title().text())
        return widget

    def clear_items(self) -> None:
        """Clear all items, freeing memory"""
        self.items = None
        for index in range(self.count()):
            self.widget(index).unset_item()
        self.clear()

    def reset(self) -> None:
        """Reset transform image settings"""
        for index in range(self.count()):
            self.widget(index).reset()

    def accept_changes(self) -> None:
        """Accept all changes"""
        self.output_arrays = []
        for index in range(self.count()):
            widget: TransfWidget = self.widget(index)
            widget.transf.accept_changes()
            self.output_arrays.append(widget.transf.output_array)

    def reject_changes(self) -> None:
        """Reject all changes"""
        for index in range(self.count()):
            self.widget(index).reject_changes()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Rotate/crop dialog
------------------

The `rotatecrop` module provides a dialog box providing essential GUI elements
for rotating (arbitrary angle) and cropping an image:

    * :py:class:`.widgets.rotatecrop.RotateCropDialog`: dialog box
    * :py:class:`.widgets.rotatecrop.RotateCropWidget`: equivalent widget

Reference
~~~~~~~~~

.. autoclass:: RotateCropDialog
   :members:
.. autoclass:: RotateCropWidget
   :members:
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from guidata.qthelpers import win32_fix_title_bar_background
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW
from qtpy.QtWidgets import QWidget  # Helping out python_qt_documentation

from plotpy.builder import make
from plotpy.config import _
from plotpy.items import get_image_in_shape
from plotpy.widgets import basetransform

if TYPE_CHECKING:
    import numpy as np

    from plotpy.items import AnnotatedRectangle, TrImageItem
    from plotpy.plot import PlotOptions
    from plotpy.plot.manager import PlotManager


class RotateCropTransform(basetransform.BaseTransform):
    """Rotate & Crop utils class, to be mixed with a class providing the
    get_plot method, like PlotDialog or RotateCropWidget (see below)

    Args:
        parent: Parent widget
        manager: Plot manager
    """

    def __init__(self, parent: RotateCropWidget, manager: PlotManager) -> None:
        super().__init__(parent, manager)
        self.crop_rect: AnnotatedRectangle = None
        self.manager = manager

    # ------BaseTransformUtils API----------------------------------------------

    def set_item(self, item: TrImageItem) -> None:
        """Set associated item

        Args:
            item: Associated item
        """
        super().set_item(item)
        crect: AnnotatedRectangle = make.annotated_rectangle(
            0, 0, 1, 1, _("Cropping rectangle")
        )
        self.crop_rect = crect
        crect.annotationparam.format = "%.1f cm"
        plot = self.manager.get_plot()
        plot.add_item(crect)
        plot.set_active_item(crect)
        x0, y0, x1, y1 = self.item.get_crop_coordinates()
        crect.set_rect(x0, y0, x1, y1)
        plot.replot()

    def reset_transformation(self) -> None:
        """Reset transformation"""
        x0, y0, x1, y1 = self.item.border_rect.get_rect()
        self.crop_rect.set_rect(x0, y0, x1, y1)

    def apply_transformation(self) -> None:
        """Apply transformation, e.g. crop or rotate"""
        # Let's crop!
        i_points = self.item.border_rect.get_points()
        xmin, ymin = i_points.min(axis=0)
        xmax, ymax = i_points.max(axis=0)
        xc0, yc0, xc1, yc1 = self.crop_rect.shape.get_rect()
        left = max([0, xc0 - xmin])
        right = max([0, xmax - xc1])
        top = max([0, yc0 - ymin])
        bottom = max([0, ymax - yc1])
        self.item.set_crop(left, top, right, bottom)
        #        print "set_crop:", left, top, right, bottom
        self.item.compute_bounds()
        self.manager.get_plot().replot()

    def compute_transformation(self) -> np.ndarray:
        """Compute transformation, return compute output array

        Returns:
            Compute output array
        """
        return get_image_in_shape(self.crop_rect, apply_interpolation=False)

    # ------Private API---------------------------------------------------------
    def show_crop_rect(self, state: bool) -> None:
        """Show/hide cropping rectangle shape

        Args:
            state: Show/hide state
        """
        self.crop_rect.setVisible(state)
        self.crop_rect.label.setVisible(state)
        self.manager.get_plot().replot()


class RotateCropDialog(QW.QDialog):
    """Rotate & Crop Dialog

    Rotate and crop a :py:class:`.TrImageItem` plot item

    Args:
        parent: Parent widget
        title: Dialog title
        options: Options dict
        resize_to: Resize dialog to (width, height)
        edit: Show "Edit" button
        toolbar: Show toolbar
    """

    def __init__(
        self,
        parent: QWidget,
        title: str | None = None,
        options: PlotOptions | dict[str, Any] | None = None,
        resize_to: tuple[int, int] | None = None,
        edit: bool = True,
        toolbar: bool = False,
    ) -> None:
        super().__init__(parent)
        win32_fix_title_bar_background(self)

        if resize_to is not None:
            width, height = resize_to
            self.resize(width, height)

        self.button_box = None

        if title is None:
            title = _("Rotate & Crop")
        self.setWindowTitle(title)
        self.widget = RotateCropWidget(parent=parent, options=options, toolbar=toolbar)
        self.setWindowFlags(QC.Qt.WindowType.Window)

        buttonhlayout = QW.QHBoxLayout()
        self.add_buttons_to_layout(buttonhlayout, edit)

        dialogvlayout = QW.QVBoxLayout()
        dialogvlayout.addWidget(self.widget.plot_widget.toolbar)
        dialogvlayout.addWidget(self.widget)
        dialogvlayout.addLayout(buttonhlayout)
        self.setLayout(dialogvlayout)

        self.transform = self.widget.transform
        self.plot_widget = self.widget.plot_widget
        self.manager = self.widget.plot_widget.manager

        self.accepted.connect(self.transform.accept_changes)
        self.rejected.connect(self.transform.reject_changes)

    def add_buttons_to_layout(self, layout: QW.QBoxLayout, edit: bool) -> None:
        """Add buttons to layout

        Args:
            layout: Layout
        """
        if edit:
            self.button_box = bbox = QW.QDialogButtonBox(
                QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel
            )
            bbox.accepted.connect(self.accept)
            bbox.rejected.connect(self.reject)
            layout.addWidget(bbox)


class RotateCropWidget(basetransform.BaseTransformWidget):
    """Rotate & Crop Widget

    Rotate and crop a :py:class:`.TrImageItem` plot item

    Args:
        parent: Parent widget
        toolbar: Show toolbar
        options: Plot options
    """

    def __init__(
        self,
        parent: QWidget,
        toolbar: bool = False,
        options: PlotOptions | dict[str, Any] | None = None,
    ) -> None:
        super().__init__(parent, toolbar=toolbar, options=options)
        self.transform = RotateCropTransform(self, self.plot_widget.manager)
        self.manager = self.plot_widget.manager

    def add_buttons_to_layout(self, layout: QW.QBoxLayout) -> None:
        """Add tool buttons to layout

        Args:
            layout: Layout
        """
        # Show crop rectangle checkbox
        show_crop = QW.QCheckBox(_("Show cropping rectangle"), self.plot_widget)
        show_crop.setChecked(True)
        show_crop.toggled.connect(self.show_crop_rect)
        layout.addWidget(show_crop)
        layout.addSpacing(15)
        basetransform.BaseTransformWidget.add_buttons_to_layout(self, layout)

    def apply_transformation(self) -> None:
        """Apply transformation"""
        self.transform.apply_transformation()

    def reset(self) -> None:
        """Reset transformation"""
        self.transform.reset()

    def reject_changes(self) -> None:
        """Reject changes"""
        self.transform.reject_changes()

    def show_crop_rect(self, state: bool) -> None:
        """Show/hide cropping rectangle shape

        Args:
            state: Show/hide state
        """
        self.transform.show_crop_rect(state)


class MultipleRotateCropWidget(basetransform.BaseMultipleTransformWidget):
    """Multiple Rotate & Crop Widget

    Rotate and crop several :py:class:`.TrImageItem` plot items"""

    TRANSFORM_WIDGET_CLASS = RotateCropWidget



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
ColorMapManager test

This plotpy widget can be used to manage colormaps (visualize, edit, create ans save).
"""

# guitest: show

import pytest
import qtpy.QtCore as QC
import qtpy.QtGui as QG
from guidata.env import execenv
from guidata.qthelpers import exec_dialog, qt_app_context

from plotpy.mathutils.colormap import delete_cmap, get_cmap
from plotpy.widgets.colormap.manager import ColorMapManager
from plotpy.widgets.colormap.widget import EditableColormap


@pytest.fixture
def test_cmap(cmap_name="Kinda Viridis"):
    cmap = EditableColormap(
        [QG.QColor(QC.Qt.GlobalColor.blue), QG.QColor(QC.Qt.GlobalColor.yellow)],
        name=cmap_name,
    )
    yield cmap

    delete_cmap(get_cmap(cmap_name))


def test_colormap_manager(test_cmap: EditableColormap) -> None:
    """Test the colormap manager widget."""
    with qt_app_context(exec_loop=False):
        red = QG.QColor(QC.Qt.GlobalColor.red)
        cmap_editor = ColorMapManager(None, active_colormap="YlGn")
        cmap_editor.show()

        with execenv.context(accept_dialogs=True):
            cmap_editor.save_colormap(test_cmap)
        cmap_editor.colormap_editor.colormap_widget.add_handle_at_relative_pos(0.5, red)
        cmap_editor.get_colormap()
        cmap_editor.colormap_editor.update_colormap_widget()
        cmap_editor.colormap_editor.update_current_dataset()

        # set the colormap to last one
        with execenv.context(accept_dialogs=True):
            cmap_editor.remove_colormap()
            result = exec_dialog(cmap_editor)

        execenv.print("Dialog result:", result)
        cmap = cmap_editor.get_colormap()
        execenv.print("Selected colormap:", None if cmap is None else cmap.name)
        delete_cmap(test_cmap)


if __name__ == "__main__":
    test_name = "Kinda Viridis"
    new_colormap = EditableColormap(
        [QG.QColor(QC.Qt.GlobalColor.blue), QG.QColor(QC.Qt.GlobalColor.yellow)],
        name=test_name,
    )
    test_colormap_manager(new_colormap)
    delete_cmap(get_cmap(test_name))



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Image file dialog helpers
-------------------------

Overview
^^^^^^^^

The :py:mod:``.imagefile`` module provides helper functions for opening and
saving image files using Qt dialogs and plotpy's I/O features.

Ready-to-use open/save dialogs:

    :py:func:`.imagefile.exec_image_save_dialog`
        Executes an image save dialog box (QFileDialog.getSaveFileName)
    :py:func:`.imagefile.exec_image_open_dialog`
        Executes an image open dialog box (QFileDialog.getOpenFileName)
    :py:func:`.imagefile.exec_images_open_dialog`
        Executes an image*s* open dialog box (QFileDialog.getOpenFileNames)

Reference
^^^^^^^^^

.. autofunction:: exec_image_save_dialog
.. autofunction:: exec_image_open_dialog
.. autofunction:: exec_images_open_dialog
"""

from __future__ import annotations

import os.path as osp
import sys

import numpy as np
from qtpy import QtWidgets as QW
from qtpy.QtWidgets import QWidget  # only to help intersphinx find QWidget

from plotpy import io
from plotpy.config import _


# ===============================================================================
# Ready-to-use open/save dialogs
# ===============================================================================
def exec_image_save_dialog(
    parent: QWidget,
    data: np.ndarray,
    template: dict | None = None,
    basedir: str = "",
    app_name: str | None = None,
) -> str | None:
    """Executes an image save dialog box (QFileDialog.getSaveFileName)

    Args:
        parent: parent widget (None means no parent)
        data: image pixel array data
        template: image template (pydicom dataset) for DICOM files
        basedir: base directory ('' means current directory)
        app_name: application name (used as a title for an eventual
         error message box in case something goes wrong when saving image)

    Returns:
        Filename if dialog is accepted, None otherwise
    """
    saved_in, saved_out, saved_err = sys.stdin, sys.stdout, sys.stderr
    sys.stdout = None
    filename, _filter = QW.QFileDialog.getSaveFileName(
        parent,
        _("Save as"),
        basedir,
        io.iohandler.get_filters("save", dtype=data.dtype, template=template),
    )
    sys.stdin, sys.stdout, sys.stderr = saved_in, saved_out, saved_err
    if filename:
        filename = str(filename)
        kwargs = {}
        if osp.splitext(filename)[1].lower() == ".dcm":
            kwargs["template"] = template
        try:
            io.imwrite(filename, data, **kwargs)
            return filename
        except Exception as msg:
            import traceback

            traceback.print_exc()
            message = _("{filename} could not be written:\n{msg}").format(
                filename=osp.basename(filename), msg=msg
            )
            QW.QMessageBox.critical(
                parent, _("Error") if app_name is None else app_name, message
            )
            return


def exec_image_open_dialog(
    parent: QWidget,
    basedir: str = "",
    app_name: str | None = None,
    to_grayscale: bool = True,
    dtype: np.dtype | None = None,
) -> tuple[str, np.ndarray] | None:
    """Executes an image open dialog box (QFileDialog.getOpenFileName)

    Args:
        parent: parent widget (None means no parent)
        basedir: base directory ('' means current directory)
        app_name: application name (used as a title for an eventual
         error message box in case something goes wrong when saving image)
        to_grayscale (bool | None): convert image to grayscale
        dtype: data type of the image

    Returns:
        (filename, data) tuple if dialog is accepted, None otherwise
    """
    saved_in, saved_out, saved_err = sys.stdin, sys.stdout, sys.stderr
    sys.stdout = None
    filename, _filter = QW.QFileDialog.getOpenFileName(
        parent, _("Open"), basedir, io.iohandler.get_filters("load", dtype=dtype)
    )
    sys.stdin, sys.stdout, sys.stderr = saved_in, saved_out, saved_err
    filename = str(filename)
    try:
        data = io.imread(filename, to_grayscale=to_grayscale)
    except Exception as msg:
        import traceback

        traceback.print_exc()
        message = _("{filename} could not be opened:\n{msg}").format(
            filename=osp.basename(filename), msg=msg
        )
        QW.QMessageBox.critical(
            parent, _("Error") if app_name is None else app_name, message
        )
        return
    return filename, data


def exec_images_open_dialog(
    parent: QWidget,
    basedir: str = "",
    app_name: str | None = None,
    to_grayscale: bool = True,
    dtype: np.dtype | None = None,
) -> list[tuple[str, np.ndarray]] | None:
    """Executes an image*s* open dialog box (QFileDialog.getOpenFileNames)

    Args:
        parent: parent widget (None means no parent)
        basedir: base directory ('' means current directory)
        app_name: application name (used as a title for an eventual
         error message box in case something goes wrong when saving image)
        to_grayscale (bool | None): convert image to grayscale
        dtype: data type

    Returns:
        (filename, data) tuples if dialog is accepted, None otherwise
    """
    saved_in, saved_out, saved_err = sys.stdin, sys.stdout, sys.stderr
    sys.stdout = None
    filenames, _filter = QW.QFileDialog.getOpenFileNames(
        parent, _("Open"), basedir, io.iohandler.get_filters("load", dtype=dtype)
    )
    sys.stdin, sys.stdout, sys.stderr = saved_in, saved_out, saved_err
    filenames = [str(fname) for fname in list(filenames)]
    for filename in filenames:
        try:
            data = io.imread(filename, to_grayscale=to_grayscale)
        except Exception as msg:
            import traceback

            traceback.print_exc()
            message = _("{filename} could not be opened:\n{msg}").format(
                filename=osp.basename(filename), msg=msg
            )
            QW.QMessageBox.critical(
                parent, _("Error") if app_name is None else app_name, message
            )
            return
        yield filename, data



# -*- coding: utf-8 -*-

from __future__ import annotations

import math
import sys
from typing import TYPE_CHECKING

import numpy as np
from guidata.utils.misc import assert_interfaces_valid
from qtpy import QtCore as QC
from qtpy import QtGui as QG
from qwt import QwtSymbol

from plotpy.items.shape.polygon import PolygonShape
from plotpy.mathutils.geometry import compute_angle, compute_center

if TYPE_CHECKING:
    import qwt.scale_map
    from qtpy.QtCore import QLineF, QPointF, QRectF
    from qtpy.QtGui import QPainter, QPolygonF

    from plotpy.styles.shape import ShapeParam


class EllipseShape(PolygonShape):
    """Ellipse shape

    Args:
        x1: X coordinate of the first point
        y1: Y coordinate of the first point
        x2: X coordinate of the second point
        y2: Y coordinate of the second point
        shapeparam: Shape parameters
    """

    CLOSED = True
    _icon_name = "circle.png"

    def __init__(
        self,
        x1: float = 0.0,
        y1: float = 0.0,
        x2: float = 0.0,
        y2: float = 0.0,
        shapeparam: ShapeParam = None,
    ) -> None:
        super().__init__(shapeparam=shapeparam)
        self.is_ellipse = False
        self.set_xdiameter(x1, y1, x2, y2)

    def switch_to_ellipse(self):
        """Switch to ellipse mode"""
        self.is_ellipse = True
        self.set_icon_name("ellipse_shape.png")

    def switch_to_circle(self):
        """Switch to circle mode"""
        self.is_ellipse = False
        self.set_icon_name("circle.png")

    def set_xdiameter(self, x0: float, y0: float, x1: float, y1: float) -> None:
        """Set the coordinates of the ellipse's X-axis diameter

        Args:
            x0: X coordinate of the first point
            y0: Y coordinate of the first point
            x1: X coordinate of the second point
            y1: Y coordinate of the second point
        """
        xline = QC.QLineF(x0, y0, x1, y1)
        yline = xline.normalVector()
        yline.translate(xline.pointAt(0.5) - xline.p1())
        if self.is_ellipse:
            yline.setLength(self.get_yline().length())
        else:
            yline.setLength(xline.length())
        yline.translate(yline.pointAt(0.5) - yline.p2())
        self.set_points(
            [(x0, y0), (x1, y1), (yline.x1(), yline.y1()), (yline.x2(), yline.y2())]
        )

    def get_xdiameter(self) -> tuple[tuple[float, float], tuple[float, float]]:
        """Return the coordinates of the ellipse's X-axis diameter

        Returns:
            Tuple with two tuples of floats
        """
        return tuple(self.points[0]) + tuple(self.points[1])

    def set_ydiameter(self, x2: float, y2: float, x3: float, y3: float) -> None:
        """Set the coordinates of the ellipse's Y-axis diameter

        Args:
            x2: X coordinate of the first point
            y2: Y coordinate of the first point
            x3: X coordinate of the second point
            y3: Y coordinate of the second point
        """
        yline = QC.QLineF(x2, y2, x3, y3)
        xline = yline.normalVector()
        xline.translate(yline.pointAt(0.5) - yline.p1())
        if self.is_ellipse:
            xline.setLength(self.get_xline().length())
        xline.translate(xline.pointAt(0.5) - xline.p2())
        self.set_points(
            [(xline.x2(), xline.y2()), (xline.x1(), xline.y1()), (x2, y2), (x3, y3)]
        )

    def get_ydiameter(self) -> tuple[tuple[float, float], tuple[float, float]]:
        """Return the coordinates of the ellipse's Y-axis diameter

        Returns:
            Tuple with two tuples of floats
        """
        return tuple(self.points[2]) + tuple(self.points[3])

    def get_rect(self) -> tuple[float, float, float, float]:
        """Get the bounding rectangle of the shape

        Returns:
            Tuple with four floats

        .. warning::

            This method is valid for circle only!
        """
        assert not self.is_ellipse
        (x0, y0), (x1, y1) = self.points[0], self.points[1]
        xc, yc = 0.5 * (x0 + x1), 0.5 * (y0 + y1)
        radius = 0.5 * np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)
        return xc - radius, yc - radius, xc + radius, yc + radius

    def boundingRect(self) -> QC.QRectF:
        """Return the bounding rectangle of the shape

        Returns:
            Bounding rectangle of the shape
        """
        # See https://stackoverflow.com/a/14163413
        (x0, y0), (x1, y1) = self.points[0], self.points[1]
        radius1 = 0.5 * np.sqrt((x1 - x0) ** 2 + (y1 - y0) ** 2)

        (x2, y2), (x3, y3) = self.points[2], self.points[3]
        radius2 = 0.5 * np.sqrt((x3 - x2) ** 2 + (y3 - y2) ** 2)

        xc, yc = 0.5 * (x0 + x1), 0.5 * (y0 + y1)
        phi = math.radians(compute_angle(xc, yc, x1, y1))

        ux = radius1 * math.cos(phi)
        uy = radius1 * math.sin(phi)
        vx = radius2 * math.cos(phi + math.pi / 2)
        vy = radius2 * math.sin(phi + math.pi / 2)

        bbox_halfwidth = math.sqrt(ux * ux + vx * vx)
        bbox_halfheight = math.sqrt(uy * uy + vy * vy)

        return QC.QRectF(
            xc - bbox_halfwidth,
            yc - bbox_halfheight,
            2 * bbox_halfwidth,
            2 * bbox_halfheight,
        )

    def get_center(self) -> tuple[float, float]:
        """Return center coordinates

        Returns:
            Tuple with two floats (x, y)
        """
        return compute_center(*self.get_xdiameter())

    def set_rect(self, x0: float, y0: float, x1: float, y1: float) -> None:
        """Set the bounding rectangle of the shape

        Args:
            x0: X coordinate of the first point
            y0: Y coordinate of the first point
            x1: X coordinate of the second point
            y1: Y coordinate of the second point

        .. warning::

            This method is valid for circle only!
        """
        assert not self.is_ellipse
        self.set_xdiameter(x0, 0.5 * (y0 + y1), x1, 0.5 * (y0 + y1))

    def compute_elements(
        self, xMap: qwt.scale_map.QwtScaleMap, yMap: qwt.scale_map.QwtScaleMap
    ) -> tuple[QPolygonF, QLineF, QLineF, QRectF]:
        """Return points, lines and ellipse rect

        Args:
            xMap: X axis scale map
            yMap: Y axis scale map

        Returns:
            Tuple with four elements (points, line0, line1, rect)
        """
        points = self.transform_points(xMap, yMap)
        line0 = QC.QLineF(points[0], points[1])
        line1 = QC.QLineF(points[2], points[3])
        rect = QC.QRectF()
        rect.setWidth(line0.length())
        rect.setHeight(line1.length())
        rect.moveCenter(line0.pointAt(0.5))
        return points, line0, line1, rect

    def hit_test(self, pos: QPointF) -> tuple[float, float, bool, None]:
        """Return a tuple (distance, attach point, inside, other_object)

        Args:
            pos: Position

        Returns:
            tuple: Tuple with four elements: (distance, attach point, inside,
             other_object).

        Description of the returned values:

        * distance: distance in pixels (canvas coordinates) to the closest
           attach point
        * attach point: handle of the attach point
        * inside: True if the mouse button has been clicked inside the object
        * other_object: if not None, reference of the object which will be
           considered as hit instead of self
        """
        if not self.plot():
            return sys.maxsize, 0, False, None
        dist, handle, inside, other = self.poly_hit_test(
            self.plot(), self.xAxis(), self.yAxis(), pos
        )
        if not inside:
            xMap = self.plot().canvasMap(self.xAxis())
            yMap = self.plot().canvasMap(self.yAxis())
            _points, _line0, _line1, rect = self.compute_elements(xMap, yMap)
            inside = rect.contains(QC.QPointF(pos))
        return dist, handle, inside, other

    def draw(
        self,
        painter: QPainter,
        xMap: qwt.scale_map.QwtScaleMap,
        yMap: qwt.scale_map.QwtScaleMap,
        canvasRect: QRectF,
    ) -> None:
        """Draw the item

        Args:
            painter: Painter
            xMap: X axis scale map
            yMap: Y axis scale map
            canvasRect: Canvas rectangle
        """
        points, line0, line1, rect = self.compute_elements(xMap, yMap)
        pen, brush, symbol = self.get_pen_brush(xMap, yMap)
        painter.setRenderHint(QG.QPainter.Antialiasing)
        painter.setPen(pen)
        painter.setBrush(brush)

        # Draw the axes lines connecting handles
        painter.drawLine(line0)
        painter.drawLine(line1)

        # For the ellipse, we need to handle non-uniform aspect ratios properly.
        # The ellipse should have its semi-axes aligned with line0 and line1 directions.
        # We use QPainterPath to draw an ellipse transformed to match the actual
        # geometry.
        center = line0.pointAt(0.5)

        # Create the ellipse in a local coordinate system where line0 is horizontal
        # and line1 is vertical, then transform it to match the actual geometry
        path = QG.QPainterPath()

        # Calculate the angle between line0 and line1 in pixel space
        # (they should be 90° in data space but may differ after transformation)
        angle0 = math.radians(line0.angle())
        angle1 = math.radians(line1.angle())

        # Semi-axes lengths
        a = line0.length() / 2  # semi-major axis (along line0)
        b = line1.length() / 2  # semi-minor axis (along line1)

        # Draw ellipse using parametric form, accounting for non-perpendicular axes
        # We sample points around the ellipse and create a path
        n_points = 72  # Number of points for smooth ellipse
        first_point = True
        for i in range(n_points + 1):
            t = 2 * math.pi * i / n_points
            # Parametric ellipse with potentially non-perpendicular axes
            # Point = center + cos(t) * a * dir0 + sin(t) * b * dir1
            dx = math.cos(t) * a * math.cos(angle0) + math.sin(t) * b * math.cos(angle1)
            dy = -math.cos(t) * a * math.sin(angle0) - math.sin(t) * b * math.sin(
                angle1
            )
            px = center.x() + dx
            py = center.y() + dy
            if first_point:
                path.moveTo(px, py)
                first_point = False
            else:
                path.lineTo(px, py)
        path.closeSubpath()
        painter.drawPath(path)

        if symbol != QwtSymbol.NoSymbol:
            for i in range(points.size()):
                symbol.drawSymbol(painter, points[i].toPoint())

    def get_xline(self) -> QC.QLineF:
        """Get the X axis diameter

        Returns:
            X axis diameter
        """
        return QC.QLineF(*(tuple(self.points[0]) + tuple(self.points[1])))

    def get_yline(self) -> QC.QLineF:
        """Get the Y axis diameter

        Returns:
            Y axis diameter
        """
        return QC.QLineF(*(tuple(self.points[2]) + tuple(self.points[3])))

    def move_point_to(
        self, handle: int, pos: tuple[float, float], ctrl: bool = False
    ) -> None:
        """Move a handle as returned by hit_test to the new position

        Args:
            handle: Handle
            pos: Position
            ctrl: True if <Ctrl> button is being pressed, False otherwise
        """
        nx, ny = pos
        if handle == 0:
            x1, y1 = self.points[1]
            if ctrl:
                # When <Ctrl> is pressed, the center position is unchanged
                x0, y0 = self.points[0]
                x1, y1 = x1 + x0 - nx, y1 + y0 - ny
            self.set_xdiameter(nx, ny, x1, y1)
        elif handle == 1:
            x0, y0 = self.points[0]
            if ctrl:
                # When <Ctrl> is pressed, the center position is unchanged
                x1, y1 = self.points[1]
                x0, y0 = x0 + x1 - nx, y0 + y1 - ny
            self.set_xdiameter(x0, y0, nx, ny)
        elif handle == 2:
            x3, y3 = self.points[3]
            if ctrl:
                # When <Ctrl> is pressed, the center position is unchanged
                x2, y2 = self.points[2]
                x3, y3 = x3 + x2 - nx, y3 + y2 - ny
            self.set_ydiameter(nx, ny, x3, y3)
        elif handle == 3:
            x2, y2 = self.points[2]
            if ctrl:
                # When <Ctrl> is pressed, the center position is unchanged
                x3, y3 = self.points[3]
                x2, y2 = x2 + x3 - nx, y2 + y3 - ny
            self.set_ydiameter(x2, y2, nx, ny)
        elif handle == -1:
            delta = (nx, ny) - self.points.mean(axis=0)
            self.points += delta

    def __reduce__(self) -> tuple:
        """Return a tuple used by pickle"""
        state = (self.shapeparam, self.points, self.z())
        return (self.__class__, (), state)

    def __setstate__(self, state: tuple) -> None:
        """Set the state of the object from a tuple"""
        self.shapeparam, self.points, z = state
        self.setZ(z)
        self.shapeparam.update_item(self)


assert_interfaces_valid(EllipseShape)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Image with custom X/Y axes linear scales"""

# guitest: show

from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import vistools as ptv


def test_imagexy():
    """Testing XYImageItem"""
    with qt_app_context(exec_loop=True):
        x, y, data = ptd.gen_xyimage()
        _win = ptv.show_items(
            [make.xyimage(x, y, data)],
            plot_type="image",
            xlabel="x (a.u.)",
            ylabel="y (a.u.)",
            wintitle=test_imagexy.__doc__,
        )


if __name__ == "__main__":
    test_imagexy()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""PlotManager test"""

# guitest: show

from guidata.qthelpers import qt_app_context, win32_fix_title_bar_background
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.builder import make
from plotpy.panels import ContrastAdjustment, PlotItemList
from plotpy.plot import BasePlot, BasePlotOptions
from plotpy.plot.manager import PlotManager
from plotpy.tests import data as ptd
from plotpy.tests import get_path


class CentralWidget(QW.QWidget):
    def __init__(self, parent):
        QW.QWidget.__init__(self, parent)

        layout = QW.QGridLayout()
        self.setLayout(layout)

        options = BasePlotOptions(type="image")
        self.plot1 = BasePlot(self, options=options)
        layout.addWidget(self.plot1, 0, 0, 1, 1)
        self.plot2 = BasePlot(self, options=options)
        layout.addWidget(self.plot2, 1, 0, 1, 1)

        self.contrast = ContrastAdjustment(self)
        layout.addWidget(self.contrast, 2, 0, 1, 2)
        self.itemlist = PlotItemList(self)
        layout.addWidget(self.itemlist, 0, 1, 2, 1)

        self.manager = PlotManager(self)
        for plot in (self.plot1, self.plot2):
            self.manager.add_plot(plot)
        for panel in (self.itemlist, self.contrast):
            self.manager.add_panel(panel)

    def register_tools(self):
        self.manager.register_all_image_tools()

    def minimumSizeHint(self):
        """Override QWidget method"""
        # The purpose of this reimplementation is to trigger the issue #3
        # (see https://github.com/PlotPyStack/PlotPy/issues/3)
        return QC.QSize(150, 100)


class Window(QW.QMainWindow):
    def __init__(self):
        super().__init__()
        win32_fix_title_bar_background(self)
        self.setWindowTitle("PlotManager test")

        filename = get_path("brain.png")
        image1 = make.image(filename=filename, title="Original", colormap="gray")
        image2 = make.image(ptd.gen_image4(500, 500))

        widget = CentralWidget(self)
        self.setCentralWidget(widget)

        widget.plot1.add_item(image1)
        widget.plot2.add_item(image2)

        toolbar = self.addToolBar("tools")
        widget.manager.add_toolbar(toolbar, id(toolbar))
        #        widget.manager.set_default_toolbar(toolbar)
        widget.register_tools()


def test_manager():
    """Test"""
    with qt_app_context(exec_loop=True):
        win = Window()
        win.show()


def test_resize_to_minimum_width():
    """Testing resize window to minimum width"""
    # This test was failing at the time of Issue #3
    # (see https://github.com/PlotPyStack/PlotPy/issues/3)
    with qt_app_context(exec_loop=False):
        win = Window()
        win.show()
        width = 350
        while width > win.minimumWidth():
            width -= 1
            win.resize(width, 100)


if __name__ == "__main__":
    test_manager()
    # test_resize_to_minimum_width()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Colormap widget
---------------

This module provides a basic widget to edit a colormap that contains a multi-slider
and a colorap representation.


Reference
~~~~~~~~~

.. autoclass:: EditableColormap
    :members:
.. autoclass:: ColorMapWidget
    :members:
"""

from __future__ import annotations

from typing import Sequence, Tuple, TypeVar, Union

import numpy as np
import qtpy.QtCore as QC
import qtpy.QtGui as QG
import qtpy.QtWidgets as QW
from guidata.qthelpers import add_actions, create_action
from qwt import QwtInterval, QwtLinearColorMap, toQImage
from qwt.color_map import ColorStop

from plotpy.config import _
from plotpy.widgets.colormap._slider import QColorMapSlider

#: Type variable for color types
ColorTypeT = TypeVar("ColorTypeT", bound=Union[QG.QColor, int])

#: Type for color initialization types
QColorInitTypes = Union[
    QG.QColor, int, Tuple[int, int, int, int], str, QC.Qt.GlobalColor, None
]


class EditableColormap(QwtLinearColorMap):
    """Overload of the QwtLinearColorMap class to add some features. This class is
    temporary and should be removed when its features are added to QwtPython.

    Args:
        *args: QwtLinearColorMap arguments
        name: Optional str name given to the colormap. Useful for the interactions
         with the rest of PlotPy, notably with the gobal colormaps dictionaries
         such as colormaps.ALL_COLORMAPS. If None, the colormap name will be set to
         "temporary". Defaults to None.
    """

    def __init__(self, *args, name: str | None = None) -> None:
        super().__init__(*args)
        self.name = name or "temporary"
        self.invert = False

    def __deepcopy__(self, memo: dict[int, object]) -> EditableColormap:
        """Deepcopy method to copy the colormap object.

        Args:
            memo: dict to store the copied objects

        Returns:
            New EditableColormap instance.
        """
        new_instance = EditableColormap(self.color1(), self.color2(), name=self.name)
        for stop in self.colorStops():
            stop: ColorStop
            new_instance.addColorStop(stop.pos, stop.rgb)
        return new_instance

    def rgb(self, interval: QwtInterval, value: float) -> int:
        """Returns the color of the colormap at the given value.
        This overriden method is used to add the possibility to invert the colormap
        colors.

        Args:
            interval: QwtInterval of the colormap
            value: value to get the color from

        Returns:
            Color of the colormap at the given value.
        """
        if self.invert:
            return super().rgb(
                interval, interval.maxValue() - value + interval.minValue()
            )
        return super().rgb(interval, value)

    @property
    def color_stop_values(self) -> list[float]:
        """Returns the position values of the color stops.

        Returns:
            List of color stop positions.
        """
        return [stop.pos for stop in self.colorStops()]

    def is_boundary_stop_index(self, stop_index: int) -> bool:
        """Checks if the given index is a boundary index (first or last).

        Args:
            stop_index: stop color index to check

        Returns:
            True if the index is a boundary index, False otherwise.
        """
        return not (
            self._can_update_next_stop(stop_index)
            and self._can_update_previous_stop(stop_index)
        )

    def _can_update_next_stop(self, stop_index: int) -> bool:
        """Checks if the next color stop can be updated.

        Args:
            stop_index: current color stop index

        Returns:
            True if the next color stop can be updated, False otherwise.
        """
        return stop_index < (len(self.colorStops()) - 1)

    def _can_update_previous_stop(self, stop_index: int) -> bool:
        """Checks if the previous color stop can be updated.

        Args:
            stop_index: current color stop index

        Returns:
            True if the previous color stop can be updated, False otherwise.
        """
        return stop_index > 0

    def update_stops_steps(self, stop_index: int) -> None:
        """Updates the steps of the previous and next color stops of the given index.
        The steps are the distance between the current stop and the next color stop.
        This method can update the given color stop index, the previous one and the next
        stop if necessary.

        Args:
            stop_index: current color stop index to update.
        """
        current_stop = self.colorStops()[stop_index]
        if self._can_update_previous_stop(stop_index):
            prev_stop: ColorStop = self.colorStops()[stop_index - 1]
            prev_stop.updateSteps(current_stop)
        if self._can_update_next_stop(stop_index):
            next_stop: ColorStop = self.colorStops()[stop_index + 1]
            current_stop.updateSteps(next_stop)

    @classmethod
    def from_iterable(
        cls, iterable: Sequence[tuple[float, str]], name=None
    ) -> "EditableColormap":
        """Converts the given iterable of tuples to a colormap (instance of Self).
        The iterable must be at least of length 2. If the iterable is of length 1,
        the colormap will be composed of two identical colors. If the iterable is
        empty, the colormap will be composed of two default colors (blue and yellow).

        Args:
            iterable: Iterable of tuples composed of a float position and a hex color
            string.
            name: Name of the new colormap instance. Defaults to None.

        Returns:
            New CustomQwtLinearColormap instance.
        """
        if len(iterable) == 1:
            iterable = (iterable[0],) * 2
        elif len(iterable) == 0:
            iterable = (
                (0.0, QG.QColor(QC.Qt.GlobalColor.blue)),
                (1.0, QG.QColor(QC.Qt.GlobalColor.yellow)),
            )

        color_min, color_max = iterable[0][1], iterable[-1][1]
        colormap = cls(color_min, color_max, name=name)

        pos_min, pos_max = iterable[0][0], iterable[-1][0]
        for position, color in iterable[1:-1]:
            scaled_position = (position - pos_min) / pos_max
            colormap.addColorStop(scaled_position, color)
        return colormap

    def to_tuples(self) -> tuple[tuple[float, str], ...]:
        """Converts a colormap to a tuple of tuples composed of a float position and a
        hex color string.

        Returns:
            Tuple of tuples composed of a float position and a hex color string.
        """
        return tuple(
            (stop.pos, QG.QColor(stop.rgb).name()) for stop in self.colorStops()
        )

    def move_color_stop(
        self, stop_index: int, new_pos: float, new_color: QG.QColor | None = None
    ) -> None:
        """Moves a color stop to a new position and updates the steps of the previous
        and next color stops if necessary. Mutates the colormap object!

        Args:
            stop_index: color stop index to move
            new_pos: new color stop position
            new_color: new color stop color. If None, will use the current stop color.
             Defaults to None.
        """
        try:
            stop: ColorStop | None = self.colorStops()[stop_index]
        except IndexError as e:
            print(e)
            return

        if stop is None:
            print("No Colorstops set, aborting edition.")
            return

        new_color = new_color or QG.QColor(stop.rgb)
        new_stop = ColorStop(new_pos, new_color)
        self.colorStops()[stop_index] = new_stop
        self.update_stops_steps(stop_index)

    def delete_stop(self, stop_index: int) -> None:
        """Deletes the color stop at given index and updates the previous color stop
        steps if necessary. Mutates the colormap object!

        Args:
            stop_index: color stop index to delete
        """
        if not self.is_boundary_stop_index(stop_index):
            cstops: list[ColorStop] = self.colorStops()
            cstops.pop(stop_index)
            cstops[stop_index - 1].updateSteps(cstops[stop_index])

    def get_stop_color(self, index: int) -> QG.QColor:
        """Returns the color of the given color stop index.

        Args:
            index: color stop index

        Returns:
            Color of the given color stop index.
        """
        index = min(index, len(self.colorStops()))
        cstop: ColorStop = self.colorStops()[index]
        return QG.QColor(cstop.rgb)


class ColorMapWidget(QW.QWidget):
    """Simple colormap widget containing a horizontal slider and a colorbar image.
    The slider is used to edit the colormap that changes in real time. It is
    possible to add/remove handles by right-clicking on the colorbar image. This
    widget does not provide a way to change a handle color. For this, use the
    ColorMapEditor widget or the ColorMapManager widget.

    Args:
        parent: parent widget
        cmap_width: minimum width of the widget. Defaults to 400.
        cmap_height: minimum height of the colorbar. Defaults to 50.
        color1: first color. Ignored if the 'colormap' argument is used. If None,
         default color is blue. Defaults to None.
        color2: second color. Ignored if the 'colormap' argument is used. If None,
         default color is yellow. Defaults to None.
        colormap: colormap instance. If None, color1 and color2 will be used
         to create a new colormap. Defaults to None.
    """

    COLORMAP_CHANGED = QC.Signal()  # type: ignore
    HANDLE_SELECTED = QC.Signal(int)  # type: ignore
    HANDLE_ADDED = QC.Signal(int, float)  # type: ignore
    HANDLE_DELETED = QC.Signal(int)  # type: ignore

    def __init__(
        self,
        parent: QW.QWidget | None,
        cmap_width: int = 400,
        cmap_height: int = 50,
        color1: QG.QColor | None = None,
        color2: QG.QColor | None = None,
        colormap: EditableColormap | None = None,
    ) -> None:
        super().__init__(parent)

        self.cmap_width = cmap_width
        self.cmap_height = cmap_height
        self.min, self.max = 0.0, 1.0
        self._tolerance = 0.05
        self._px_cursor_offset = 0  # This is a patch because the cursor position is
        # not exactly the handle position...
        self._handle_selected = None

        self.slider_menu = self.setup_menu()

        multi_range_hslider = QColorMapSlider(QC.Qt.Orientation.Horizontal)
        multi_range_hslider.setTickPosition(QW.QSlider.TickPosition.TicksAbove)
        multi_range_hslider.setBarVisible(False)
        multi_range_hslider.setMinimum(self.min)
        multi_range_hslider.setMaximum(self.max)
        multi_range_hslider.setMaximumHeight(20)
        self.multi_range_hslider = multi_range_hslider
        self.set_handles_values((self.min, self.max))
        self.setContextMenuPolicy(QC.Qt.ContextMenuPolicy.CustomContextMenu)

        self._last_selection: tuple[int, int] = 0, 0

        self.qwt_color_interval = QwtInterval(self.min, self.max)

        if colormap is None:
            color1 = QG.QColor(QC.Qt.GlobalColor.blue) if color1 is None else color1
            color2 = QG.QColor(QC.Qt.GlobalColor.yellow) if color2 is None else color2
            self._colormap = EditableColormap(color1, color2)
        else:
            self._colormap = colormap
            self.set_handles_values(colormap.color_stop_values)

        self.colortable = self._colormap.colorTable(self.qwt_color_interval)

        colormap_label = QW.QLabel("Stylish colormap widget")
        colormap_label.setScaledContents(True)
        self._colormap_label = colormap_label
        self.draw_colormap_image()

        layout = QW.QVBoxLayout(self)
        layout.addWidget(colormap_label)
        layout.addWidget(multi_range_hslider)
        self.range_layout = layout

        self.setLayout(layout)
        self.setToolTip(_("Right click to add/remove a color"))

        self.COLORMAP_CHANGED.connect(self.draw_colormap_image)
        self.multi_range_hslider.valueChanged.connect(
            self._edit_color_map_on_slider_change
        )
        self.multi_range_hslider.sliderPressed.connect(self.emit_handle_selected)
        self.multi_range_hslider.sliderReleased.connect(self.release_handle)
        self.customContextMenuRequested.connect(self.open_slider_menu)

    def set_colormap(self, colormap: EditableColormap) -> None:
        """Replaces the current colormap.

        Args:
            colormap: replacement colormap
        """
        self._colormap = colormap
        self.set_handles_values(colormap.color_stop_values)
        self.COLORMAP_CHANGED.emit()

    def get_colormap(self) -> EditableColormap:
        """Get the current colormap being edited.

        Returns:
            current colormap
        """
        return self._colormap

    def update_color_table(self) -> None:
        """Updates and caches the current color table."""
        self.colortable[:] = self._colormap.colorTable(self.qwt_color_interval)

    def emit_handle_selected(self) -> None:
        """When called, computes a slider position from the current position of the
        mouse and maps it to slider range. Then finds which handles was selected and
        emits the custom signal HANDLE_SELECTED(handle_index) and caches the index."""
        self._handle_selected = self.multi_range_hslider.pressed_index
        self.HANDLE_SELECTED.emit(self._handle_selected)

    def release_handle(self) -> None:
        """When called, unset the cached current handle index."""
        self._handle_selected = None

    def draw_colormap_image(self) -> None:
        """Recomputes and redraws the colorbar image."""
        # TODO: check how to update the image without having to recreate one,
        # just by updating the colortable
        self.update_color_table()
        self._colormap_label.setPixmap(self.cmap_to_qimage())

    def get_color_from_value(
        self, value: float, as_type: type[ColorTypeT] = QG.QColor
    ) -> ColorTypeT:
        """Returns the color assigned to a given value using the current colormap.

        Args:
            value: value to assign a color to
            as_type: Color type to be returned (int or QG.QColor). Defaults to
            QG.QColor.

        Returns:
            The assigned color for the input value using the current colormap.
        """
        int_color = self._colormap.rgb(self.qwt_color_interval, value)
        return as_type(int_color)

    def get_handle_color(
        self, handle_index: int, as_type: type[ColorTypeT] = QG.QColor
    ) -> ColorTypeT:
        """Returns the color of a handled using its index. Directly return the color of
        the ColorStop object of the given index.

        Args:
            handle_index: index of the handled to assign a color too.
            as_type: Color type to be returned (int or QG.QColor). Defaults to
            QG.QColor.

        Returns:
            The assigned color for the input handle index using the current colormap.
        """
        int_color = self._colormap.get_stop_color(handle_index)
        return as_type(int_color)

    def get_hex_color(self, handle_index: int) -> str:
        """Same as self.get_handle_color but returns the color as a hex color string.

        Args:
            handle_index: index of the handled to assign a color too.

        Returns:
            Handle's hex color string
        """
        return self.get_handle_color(handle_index, QG.QColor).name()

    def edit_color_stop(
        self,
        index: int,
        new_pos: float | None = None,
        new_color: QG.QColor | int | None = None,
    ) -> None:
        """Edit an existing color stop in the current colormap. Mutates the colormap
        object. Also edits the slider handle position.

        Args:
            index: color stop index to mutate
            new_pos: new color stop position. If not set will remain the same.
             Defaults to None.
            new_color: new color. If not set will remain the same. Defaults to None.
        """
        if new_pos is None:
            new_pos = self.get_handles_tuple()[index]

        if isinstance(new_color, int):
            new_color = QG.QColor(self.colortable[new_color])

        self._colormap.move_color_stop(index, new_pos, new_color)
        self.set_handles_values(self._colormap.color_stop_values)
        self.COLORMAP_CHANGED.emit()

    def _edit_color_map_on_slider_change(self, raw_values: tuple[float, ...]) -> None:
        """Modifies the colormap depending on the currently selected handle. Tiggered on
        hanle movement.

        Args:
            raw_values: tuple of handles position
        """
        handle_index = self._handle_selected
        values = self.multi_range_hslider.filter_first_last_positions(raw_values)
        if handle_index is not None:
            handle_pos = values[handle_index]
            if handle_index == 0 or handle_index == len(values):
                return

            self._colormap.move_color_stop(handle_index, handle_pos)
            self.COLORMAP_CHANGED.emit()

    def _get_closest_handle_index(self, pos: float) -> tuple[int, float]:
        """Return the index of the closest handle from the given position

        Args:
            pos: position along the slider axis and coordinate space

        Returns:
            index and distance of the closest point (distance can be negative)
        """
        values = self.get_handles_tuple()
        closest_handle = min(range(len(values)), key=lambda i: abs(values[i] - pos))
        return closest_handle, abs(pos - values[closest_handle])

    def _get_neighbour_positions(self, pos: float) -> tuple[float, float]:
        """Return the positions of the previous and next handles from the given
        position. If no handle is found, the method can return self.min and/or self.max

        Args:
            pos: position from which to look for the previous and next handles

        Returns:
            previous and next handle positions
        """
        values = self.get_handles_tuple()
        previous_pos = max(
            filter(lambda val: val <= pos, values[::-1]), default=self.min
        )
        next_pos = min(filter(lambda val: val > pos, values), default=self.max)

        return previous_pos, next_pos

    def _new_available_pos(self, pos: float) -> float:
        """Computes the most position to insert a new handle. This method takes into
        account the input position, the prefered offset if handles are too close,
        previous and next handles from the current position to output the best position
        possible.

        Args:
            pos: requested position

        Returns:
            Best position for insertion according to parameters cited above.
        """
        previous_pos, next_pos = self._get_neighbour_positions(pos)
        if (neighbour_diff := next_pos - previous_pos) < self._tolerance * 2:
            return previous_pos + (neighbour_diff / 2)

        if (pos - previous_pos) < self._tolerance:
            return previous_pos + self._tolerance

        if (next_pos - pos) < self._tolerance:
            return next_pos - self._tolerance

        return pos

    def _add_handle_at_click(self) -> None:
        """Adds a new handle at last right-clicked position thanks to the attribute
        self._last_selection. Mutates the colormap object.
        """
        x, _ = self._last_selection
        new_handle_value = self.multi_range_hslider.widget_pos_to_value(x)
        values = self.get_handles_list()

        new_handle_value = self._new_available_pos(new_handle_value)
        new_color = self.get_color_from_value(new_handle_value)
        self._colormap.addColorStop(new_handle_value, new_color)
        values.append(new_handle_value)
        values.sort()
        self.set_handles_values(values)
        self.HANDLE_ADDED.emit(values.index(new_handle_value), new_handle_value)

    def _delete_handle_at_click(self) -> None:
        """Tries to delete an existing handle at the right-clicked position thanks to
        the attribute self._last_selection. Does nothing if the click is too far from
        a handle. Mutates the colormap object.
        """
        x, _ = self._last_selection
        values: list[float] = self.get_handles_list()
        x_handle_coor_space = self.multi_range_hslider.widget_pos_to_value(x)
        closest_handle_index, click_distance = self._get_closest_handle_index(
            x_handle_coor_space
        )
        if (
            abs(click_distance) > self._tolerance
            or len(values) <= 2
            or closest_handle_index in (0, len(values) - 1)
        ):
            return
        values.pop(closest_handle_index)
        self.set_handles_values(tuple(values))
        self._colormap.delete_stop(closest_handle_index)

        self.HANDLE_DELETED.emit(closest_handle_index)
        self.COLORMAP_CHANGED.emit()

    def setup_menu(self) -> QW.QMenu:
        """Setups the contextual menu used to insert and delete handles.

        Returns:
            Colorbar contextual menu
        """
        new_handle_action = create_action(
            self,
            title=_("Add new color"),
            icon=None,
            triggered=self._add_handle_at_click,
        )
        delete_handle_action = create_action(
            self,
            title=_("Remove color"),
            icon=None,
            triggered=self._delete_handle_at_click,
        )
        actions = (new_handle_action, delete_handle_action)

        menu = QW.QMenu(self)
        add_actions(menu, actions)
        return menu

    def add_handle_at_relative_pos(
        self, relative_pos: float, new_color: QG.QColor | int | None = None
    ) -> None:
        """Insert a handle in the widget at the relative position (between 0. and 1.).
        Mutates the colormap object. If the relative position is already occupied by a
        handle, the new handle will be inserted at the closest available position then
        will be moved back to the requested position.

        Args:
            relative_pos: insertion position
        """
        if new_color is None:
            new_color = self._colormap.color(self.qwt_color_interval, relative_pos)

        new_relative_pos = self._new_available_pos(relative_pos)
        self._colormap.addColorStop(new_relative_pos, new_color)

        values = self.get_handles_list()
        values.append(new_relative_pos)
        values.sort()
        new_value_index = values.index(new_relative_pos)

        self.edit_color_stop(new_value_index, relative_pos, None)
        self.HANDLE_ADDED.emit(new_value_index, relative_pos)

    def get_handles_count(self) -> int:
        """Number of slider handles.

        Returns:
            handles count
        """
        return len(self.multi_range_hslider.value())

    def get_handles_list(self) -> list[float]:
        """Return the current handles as a mutable list. If the data doesn't need to be
        mutated, prefer the method self.get_handles_tuple() that directly returns the
        handles tuple without making a conversion to a list.

        Returns:
            Mutable list of handles position
        """
        return list(self.multi_range_hslider.value())

    def get_handles_tuple(self) -> tuple[float, ...]:
        """Passthrough to get the tuple of handles position.

        Returns:
            Immutable tuple of handles position
        """
        return self.multi_range_hslider.value()

    def set_handles_values(self, values: Sequence[float]) -> None:
        """Passthrough to set handles position.

        Args:
            values: sequence of values with a minimum length of 2 (min and max).
        """
        self.multi_range_hslider.setValue(values)

    def cmap_to_qimage(self) -> QG.QPixmap:
        """Build the horizontal colorab pixmap from the current colormap."""
        data = np.zeros((self.cmap_height, self.cmap_width), np.uint8)
        line = np.linspace(0, 255, self.cmap_width)
        data[:, :] = line[np.newaxis, :]
        img = toQImage(data)
        img.setColorTable(self.colortable)
        return QG.QPixmap.fromImage(img)

    def open_slider_menu(self, pos: QC.QPoint) -> None:
        """Opens the contextual menu at input position.

        Args:
            pos: contextual menu position
        """
        slider_pos = self.multi_range_hslider.mapFrom(self, pos)
        self._last_selection = (
            slider_pos.x(),
            slider_pos.y(),
        )
        glob_pos = self.mapToGlobal(pos)

        self.slider_menu.popup(glob_pos)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Contours
========

This module provides functions and classes to create contour curves.

.. autoclass:: ContourLine
   :members:

.. autofunction:: compute_contours

.. autoclass:: ContourItem
   :members:

.. autofunction:: create_contour_items
"""

from __future__ import annotations

import guidata.dataset as gds
import numpy as np
from guidata.utils.misc import assert_interfaces_valid
from skimage import measure

from plotpy.config import _
from plotpy.items.shape.polygon import PolygonShape
from plotpy.styles import ShapeParam


class ContourLine(gds.DataSet):
    """A contour line"""

    vertices = gds.FloatArrayItem(_("Vertices"), help=_("Vertices of the line"))
    level = gds.FloatItem(_("Level"), help=_("Level of the line"))


def compute_contours(
    Z: np.ndarray,
    levels: float | np.ndarray,
    X: np.ndarray | None = None,
    Y: np.ndarray | None = None,
) -> list[ContourLine]:
    """Create contour curves

    Args:
        Z: The height values over which the contour is drawn.
        levels : Determines the number and positions of the contour lines/regions.
         If a float, draw contour lines at this specified levels
         If array-like, draw contour lines at the specified levels.
         The values must be in increasing order.
        X: The coordinates of the values in *Z*.
         *X* must be 2-D with the same shape as *Z* (e.g. created via
         ``numpy.meshgrid``), or it must both be 1-D such that ``len(X) == M``
         is the number of columns in *Z*.
         If none, they are assumed to be integer indices, i.e. ``X = range(M)``.
        Y: The coordinates of the values in *Z*.
         *Y* must be 2-D with the same shape as *Z* (e.g. created via
         ``numpy.meshgrid``), or it must both be 1-D such that ``len(Y) == N``
         is the number of rows in *Z*.
         If none, they are assumed to be integer indices, i.e. ``Y = range(N)``.

    Returns:
        A list of :py:class:`ContourLine` instances.
    """
    z = np.asarray(Z, dtype=np.float64)
    if z.ndim != 2:
        raise TypeError("Input z must be a 2D array.")
    elif z.shape[0] < 2 or z.shape[1] < 2:
        raise TypeError("Input z must be at least a 2x2 array.")

    if isinstance(levels, np.ndarray):
        levels = np.asarray(levels, dtype=np.float64)
    else:
        levels = np.asarray([levels], dtype=np.float64)

    if X is None:
        delta_x, x_origin = 1.0, 0.0
    else:
        delta_x, x_origin = X[0, 1] - X[0, 0], X[0, 0]
    if Y is None:
        delta_y, y_origin = 1.0, 0.0
    else:
        delta_y, y_origin = Y[1, 0] - Y[0, 0], Y[0, 0]

    # Find contours in the binary image for each level
    clines = []
    for level in levels:
        for contour in measure.find_contours(Z, level):
            contour = contour.squeeze()
            if len(contour) > 1:  # Avoid single points
                line = np.zeros_like(contour, dtype=np.float32)
                line[:, 0] = contour[:, 1] * delta_x + x_origin
                line[:, 1] = contour[:, 0] * delta_y + y_origin
                cline = ContourLine.create(vertices=line, level=level)
                clines.append(cline)
    return clines


class ContourItem(PolygonShape):
    """Contour shape"""

    _readonly = True
    _can_select = True
    _can_resize = False
    _can_rotate = False
    _can_move = False
    _icon_name = "contour.png"

    def __init__(self, points=None, shapeparam=None):
        super().__init__(points, closed=True, shapeparam=shapeparam)


assert_interfaces_valid(ContourItem)


def create_contour_items(
    Z: np.ndarray,
    levels: float | np.ndarray,
    X: np.ndarray | None = None,
    Y: np.ndarray | None = None,
) -> list[ContourItem]:
    """Create contour items

    Args:
        Z: The height values over which the contour is drawn.
        levels : Determines the number and positions of the contour lines/regions.
         If a float, draw contour lines at this specified levels
         If array-like, draw contour lines at the specified levels.
         The values must be in increasing order.
        X: The coordinates of the values in *Z*.
         *X* must be 2-D with the same shape as *Z* (e.g. created via
         ``numpy.meshgrid``), or it must both be 1-D such that ``len(X) == M``
         is the number of columns in *Z*.
         If none, they are assumed to be integer indices, i.e. ``X = range(M)``.
        Y: The coordinates of the values in *Z*.
         *Y* must be 2-D with the same shape as *Z* (e.g. created via
         ``numpy.meshgrid``), or it must both be 1-D such that ``len(Y) == N``
         is the number of rows in *Z*.
         If none, they are assumed to be integer indices, i.e. ``Y = range(N)``.

    Returns:
        A list of :py:class:`.ContourItem` instances.
    """
    items = []

    contours = compute_contours(Z, levels, X, Y)
    for cline in contours:
        param = ShapeParam("Contour", icon="contour.png")
        item = ContourItem(points=cline.vertices, shapeparam=param)
        item.set_style("plot", "shape/contour")
        item.setTitle(_("Contour") + f"[Z={cline.level}]")
        items.append(item)
    return items



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Curve and Image Statistics tools tests"""

# guitest: show

from __future__ import annotations

from typing import TYPE_CHECKING, Type

import numpy as np
from guidata.qthelpers import create_toolbutton, execenv, qt_app_context

from plotpy.builder import make
from plotpy.config import _
from plotpy.constants import PlotType
from plotpy.plot import PlotDialog, PlotOptions
from plotpy.tests.data import gen_image4
from plotpy.tests.unit.utils import drag_mouse
from plotpy.tools import (
    CurveStatsTool,
    ImageStatsTool,
    InteractiveTool,
    YRangeCursorTool,
)

if TYPE_CHECKING:
    from plotpy.items.image.base import BaseImageItem
    from plotpy.styles import BaseImageParam


class MyPlotDialog(PlotDialog):
    def __init__(self, type: PlotType, toolclass: Type[InteractiveTool]) -> None:
        """Reimplement PlotDialog method"""
        super().__init__(
            title=f"{type.name.lower().capitalize()} Statistics tools test",
            toolbar=True,
            options=PlotOptions(type=type),
        )
        self.toolclass = toolclass
        # No need to add the tools to the manager, they are automatically added
        # when the `register_curve_tools` or `register_image_tools` method is called
        self.setup_items()
        if execenv.unattended:
            self.simulate_stats_tool()

    def populate_plot_layout(self) -> None:
        """Populate the plot layout"""
        super().populate_plot_layout()
        options = self.plot_widget.options
        btn = create_toolbutton(
            self,
            None,
            f"Simulate {options.type.name.lower()} statistics tool",
            self.simulate_stats_tool,
        )
        self.add_widget(btn)

    def simulate_stats_tool(self) -> None:
        """Simulate the statistics tool"""
        tool = self.manager.get_tool(self.toolclass)
        tool.activate()
        drag_mouse(self, np.array([0.4, 0.8]), np.array([0.2, 0.7]))

    def setup_items(self) -> None:
        """Setup items"""
        plot = self.get_plot()
        if self.plot_widget.options.type is PlotType.CURVE:
            x = np.linspace(0, 10, 100)
            y = np.sin(x)
            item = make.curve(x, y, title="Curve")
        else:
            item = make.image(gen_image4(500, 500), title="Image", colormap="cool")
        plot.add_item(item)
        plot.select_item(item)


def test_curve_x_stats_tools() -> None:
    """Test CurveStatsTool"""
    with qt_app_context(exec_loop=True):
        win = MyPlotDialog(type=PlotType.CURVE, toolclass=CurveStatsTool)
        tool = win.manager.get_tool(CurveStatsTool)
        labelfuncs = (
            ("Mean = %g", lambda *args: np.mean(args[1])),
            ("Std = %g", lambda *args: np.std(args[1])),
            ("Max = %g", lambda *args: np.max(args[1])),
            ("Min = %g", lambda *args: np.min(args[1])),
        )
        tool.set_labelfuncs(labelfuncs)
        win.show()


def test_y_range_cursor_tool() -> None:
    """Test YRangeCursorTool"""
    with qt_app_context(exec_loop=True):
        win = MyPlotDialog(type=PlotType.CURVE, toolclass=YRangeCursorTool)
        tool = win.manager.get_tool(YRangeCursorTool)
        labelfuncs = (
            ("Mean = %g", lambda *args: np.mean(args[1])),
            ("Std = %g", lambda *args: np.std(args[1])),
            ("Max = %g", lambda *args: np.max(args[1])),
            ("Min = %g", lambda *args: np.min(args[1])),
        )
        tool.set_labelfuncs(labelfuncs)
        win.show()


# Note:
# -----
#
# Using the following `get_more_stats` function, the `ImageStatsTool` will display
# the surface, the sum and the density of the selected area - as additional stats
# if the `replace` parameter is set to `False` (default value).
#
# This is the way of reimplementing the old `show_surface` and `show_integral`
# arguments of the `ImageStatsTool` class in previous versions of PlotPy.


def get_more_stats(
    item: BaseImageItem,
    x0: float,
    y0: float,
    x1: float,
    y1: float,
) -> str:
    """Return formatted string with stats on image rectangular area
    (output should be compatible with AnnotatedShape.get_info)

    Args:
        item: image item
        x0: X0
        y0: Y0
        x1: X1
        y1: Y1
    """
    ix0, iy0, ix1, iy1 = item.get_closest_index_rect(x0, y0, x1, y1)
    data = item.data[iy0:iy1, ix0:ix1]
    p: BaseImageParam = item.param
    xunit, yunit, zunit = p.get_units()
    info = ""
    if xunit == yunit:
        surfacefmt = p.xformat.split()[0] + " " + xunit
        if xunit != "":
            surfacefmt = surfacefmt + "²"
        surface = abs((x1 - x0) * (y1 - y0))
        info += _("surface = %s") % (surfacefmt % surface)
    integral = data.sum()
    integral_fmt = r"%.3e " + zunit
    info = info + "<br>" + _("sum = %s") % (integral_fmt % integral)
    if xunit == yunit and xunit is not None and zunit is not None:
        if surface != 0:
            density = integral / surface
            densityfmt = r"%.3e"
            if xunit and zunit:
                densityfmt += " " + zunit + "/" + xunit + "²"
            info = info + "<br>" + _("density = %s") % (densityfmt % density)
        else:
            info = info + "<br>" + _("density not computed : surface is null !")
    return info


def test_image_stats_tools() -> None:
    """Test"""
    with qt_app_context(exec_loop=True):
        win = MyPlotDialog(type=PlotType.IMAGE, toolclass=ImageStatsTool)
        tool = win.manager.get_tool(ImageStatsTool)
        tool.set_stats_func(get_more_stats, replace=False)
        win.show()


if __name__ == "__main__":
    test_curve_x_stats_tools()
    test_y_range_cursor_tool()
    test_image_stats_tools()



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Image filter demo"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context
from scipy.ndimage import gaussian_filter

from plotpy import io
from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import get_path


def imshow(x, y, data, filter_area, yreverse=True):
    with qt_app_context(exec_loop=True):
        win = make.dialog(
            edit=False,
            toolbar=True,
            wintitle="Image filter demo",
            xlabel="x (cm)",
            ylabel="y (cm)",
            yreverse=yreverse,
            type="image",
            size=(800, 600),
        )
        image = make.xyimage(x, y, data)
        plot = win.manager.get_plot()
        plot.add_item(image)
        xmin, xmax, ymin, ymax = filter_area

        def ifilter(x, y, data):
            """Image filter function"""
            return gaussian_filter(data, 5)

        flt = make.imagefilter(xmin, xmax, ymin, ymax, image, filter=ifilter)
        plot.add_item(flt, z=1)
        plot.replot()
        win.show()


def test_imagefilter():
    """Test image filter"""
    x, y, data = ptd.gen_xyimage()
    imshow(x, y, data, filter_area=(-3.0, -1.0, 0.0, 2.0), yreverse=False)

    filename = get_path("brain.png")
    data = io.imread(filename, to_grayscale=True)
    x = np.linspace(0, 30.0, data.shape[1])
    y = np.linspace(0, 30.0, data.shape[0])
    imshow(x, y, data, filter_area=(10.0, 20.0, 5.0, 15.0))


if __name__ == "__main__":
    test_imagefilter()



# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .annotation import (
    AnnotatedCircleTool,
    AnnotatedEllipseTool,
    AnnotatedObliqueRectangleTool,
    AnnotatedPointTool,
    AnnotatedRectangleTool,
    AnnotatedSegmentTool,
    AnnotatedPolygonTool,
)
from .axes import AxisScaleTool, PlaceAxesTool
from .base import (
    ActionTool,
    CommandTool,
    DefaultToolbarID,
    InteractiveTool,
    PanelTool,
    RectangularActionTool,
    ToggleTool,
)
from .cross_section import (
    AverageCrossSectionTool,
    CrossSectionTool,
    LCSPanelTool,
    LineCrossSectionTool,
    ObliqueCrossSectionTool,
    OCSPanelTool,
    XCSPanelTool,
    YCSPanelTool,
)
from .cursor import HCursorTool, HRangeTool, VCursorTool, XCursorTool
from .curve import (
    AntiAliasingTool,
    CurveStatsTool,
    YRangeCursorTool,
    DownSamplingTool,
    EditPointTool,
    SelectPointsTool,
    SelectPointTool,
)
from .image import (
    AspectRatioTool,
    ColormapTool,
    ContrastPanelTool,
    ImageMaskTool,
    ImageStatsTool,
    ZAxisLogTool,
    LockTrImageTool,
    OpenImageTool,
    ReverseColormapTool,
    LockLUTRangeTool,
    ReverseXAxisTool,
    ReverseYAxisTool,
    RotateCropTool,
    RotationCenterTool,
)
from .item import (
    DeleteItemTool,
    EditItemDataTool,
    ExportItemDataTool,
    ItemCenterTool,
    ItemListPanelTool,
    LoadItemsTool,
    SaveItemsTool,
)
from .label import LabelTool
from .misc import (
    AboutTool,
    CopyToClipboardTool,
    FilterTool,
    HelpTool,
    OpenFileTool,
    PrintTool,
    SaveAsTool,
    SnapshotTool,
)
from .plot import (
    BasePlotMenuTool,
    DisplayCoordsTool,
    DoAutoscaleTool,
    DummySeparatorTool,
    RectangularSelectionTool,
    RectZoomTool,
)
from .selection import SelectTool
from .shape import (
    CircleTool,
    EllipseTool,
    PolygonTool,
    MultiLineTool,
    ObliqueRectangleTool,
    PointTool,
    RectangleTool,
    RectangularShapeTool,
    SegmentTool,
)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Builder tests"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import data as ptd
from plotpy.tests import vistools as ptv


def test_builder():
    """Testing plot builder"""
    data = ptd.gen_image1()
    show = ptv.show_items
    with qt_app_context(exec_loop=True):
        _win0 = show([make.image(data, colormap="pink")], wintitle="Pink colormap test")
        _win1 = show([make.image(data)], wintitle="Default LUT range")
        img1 = make.image(data)
        img1.set_lut_range([0, 1])
        _win2 = show([img1], wintitle="0->1 LUT range through item")
        img2 = make.image(data, lut_range=[0, 1])
        _win3 = show([img2], wintitle="0->1 LUT range through builder")
        x = np.linspace(1, 10, 200)
        y = np.sin(x)

        crv1 = make.curve(x, y, dx=x * 0.1, dy=y * 0.23)
        crv2 = make.curve(x, np.cos(x), color="#FF0000")
        _win4 = show([crv1, crv2], wintitle="Error bars with make.curve()")

        dat22 = np.zeros((2, 2), np.float32)
        dat22[0, 0] = 1
        dat22[0, 1] = 2
        dat22[1, 0] = 3
        dat22[1, 1] = 4
        img3 = make.image(dat22, xdata=[-1, 3], ydata=[-1, 3], interpolation="nearest")
        _win5 = show([img3], wintitle="2x2 image")

        img4 = make.image(dat22, x=[0, 2], y=[0, 2], interpolation="nearest")
        _win6 = show([img4], wintitle="Equivalent 2x2 XY image")


if __name__ == "__main__":
    test_builder()



# -*- coding: utf-8 -*-# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Plot tools"""

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.configtools import get_icon
from guidata.qthelpers import add_actions, add_separator
from qtpy import QtCore as QC
from qtpy import QtWidgets as QW

from plotpy.config import _
from plotpy.constants import PARAMETERS_TITLE_ICON
from plotpy.events import ZoomRectHandler, setup_standard_tool_filter
from plotpy.interfaces import IImageItemType, IShapeItemType
from plotpy.items import RectangleShape, get_items_in_rectangle
from plotpy.tools.base import (
    CommandTool,
    DefaultToolbarID,
    GuiTool,
    InteractiveTool,
    RectangularActionTool,
)

if TYPE_CHECKING:
    from plotpy.plot import BasePlot, PlotManager


class DoAutoscaleTool(CommandTool):
    """A tool to perform autoscale for associated plot."""

    def __init__(
        self,
        manager,
        title=_("AutoScale"),
        icon="autoscale.png",
        tip=None,
        toolbar_id=DefaultToolbarID,
    ):
        super().__init__(
            manager, title=title, icon=icon, tip=tip, toolbar_id=toolbar_id
        )

    def setup_context_menu(self, menu: QW.QMenu, plot: BasePlot) -> None:
        """
        Set up the context menu for the tool.

        Args:
            menu: Context menu
            plot: Plot instance
        """
        pass

    def activate_command(self, plot: BasePlot, checked: bool) -> None:
        """
        Activate tool.

        Args:
            plot: Plot instance
            checked: Whether the tool is checked
        """
        if checked:
            plot.do_autoscale()


class BasePlotMenuTool(CommandTool):
    """
    A tool that gather parameter panels from the BasePlot
    and proposes to edit them and set them back
    """

    def __init__(
        self, manager, key, title=None, icon=None, tip=None, toolbar_id=DefaultToolbarID
    ):
        default_title, default_icon = PARAMETERS_TITLE_ICON[key]
        if title is None:
            title = default_title
        if icon is None:
            icon = default_icon
        super().__init__(manager, title, icon, tip, toolbar_id)
        # Warning: icon (str) --(Base class constructor)--> self.icon (QIcon)
        self.key = key

    def activate_command(self, plot: BasePlot, checked: bool) -> None:
        """
        Activate tool.

        Args:
            plot: Plot instance
            checked: Whether the tool is checked
        """
        plot.edit_plot_parameters(self.key)

    def update_status(self, plot: BasePlot) -> None:
        """
        Update the status of the tool.

        Args:
            plot: Plot instance
        """
        status = plot.get_plot_parameters_status(self.key)
        self.action.setEnabled(status)


class DisplayCoordsTool(CommandTool):
    """Tool for displaying coordinates."""

    def __init__(self, manager):
        super().__init__(
            manager,
            _("Markers"),
            icon=get_icon("on_curve.png"),
            tip=None,
            toolbar_id=None,
        )
        self.action.setEnabled(True)

    def create_action_menu(self, manager: PlotManager) -> QW.QMenu:
        """
        Create and return menu for the tool's action.

        Args:
            manager: Plot manager

        Returns:
            Menu for the tool's action
        """
        menu = QW.QMenu()
        self.canvas_act = manager.create_action(
            _("Free"), toggled=self.activate_canvas_pointer
        )
        self.curve_act = manager.create_action(
            _("Bound to active item"), toggled=self.activate_curve_pointer
        )
        add_actions(menu, (self.canvas_act, self.curve_act))
        return menu

    def activate_canvas_pointer(self, enable: bool) -> None:
        """
        Activate canvas pointer.

        Args:
            enable: Whether to enable the canvas pointer
        """
        plot = self.get_active_plot()
        if plot is not None:
            plot.set_pointer("canvas" if enable else None)

    def activate_curve_pointer(self, enable: bool) -> None:
        """
        Activate curve pointer.

        Args:
            enable: Whether to enable the curve pointer
        """
        plot = self.get_active_plot()
        if plot is not None:
            plot.set_pointer("curve" if enable else None)

    def update_status(self, plot: BasePlot) -> None:
        """
        Update the status of the tool.

        Args:
            plot: Plot instance
        """
        self.canvas_act.setChecked(plot.canvas_pointer)
        self.curve_act.setChecked(plot.curve_pointer)


class RectZoomTool(InteractiveTool):
    TITLE = _("Rectangle zoom")
    ICON = "magnifier.png"

    def setup_filter(self, baseplot):
        """

        :param baseplot:
        :return:
        """
        filter = baseplot.filter
        start_state = filter.new_state()
        handler = ZoomRectHandler(
            filter, QC.Qt.MouseButton.LeftButton, start_state=start_state
        )
        shape, h0, h1 = self.get_shape()
        handler.set_shape(shape, h0, h1)
        return setup_standard_tool_filter(filter, start_state)

    def get_shape(self):
        """

        :return:
        """
        shape = RectangleShape(0, 0, 1, 1)
        shape.set_style("plot", "shape/rectzoom")
        return shape, 0, 2


class DummySeparatorTool(GuiTool):
    """ """

    def __init__(self, manager, toolbar_id=DefaultToolbarID):
        super().__init__(manager, toolbar_id)

    def setup_toolbar(self, toolbar):
        """Setup tool's toolbar"""
        add_separator(toolbar)

    def setup_context_menu(self, menu, plot):
        """

        :param menu:
        :param plot:
        """
        add_separator(menu)


class RectangularSelectionTool(RectangularActionTool):
    SWITCH_TO_DEFAULT_TOOL = True
    TITLE = _("Rectangular selection tool")
    ICON = "select_area.svg"

    def __init__(self, manager, intersect=True, toolbar_id=DefaultToolbarID):
        super().__init__(
            manager, self.select_items, toolbar_id=toolbar_id, fix_orientation=True
        )
        self.intersect = intersect

    def select_items(self, plot, p0, p1):
        items_to_select = []
        # select items that implement IShapeItemType (annotation shapes, ...)
        items = get_items_in_rectangle(
            plot,
            p0,
            p1,
            item_type=IShapeItemType,
            intersect=self.intersect,
        )
        for item in items:
            if item.isVisible():
                items_to_select.append(item)
        # select items that implement IExportROIImageItemType (TrImageItem, ...)
        items = get_items_in_rectangle(
            plot,
            p0,
            p1,
            item_type=IImageItemType,
            intersect=self.intersect,
        )
        for item in items:
            if item.isVisible():
                items_to_select.append(item)
        plot.select_some_items(items_to_select)



# -*- coding: utf-8 -*-

# pylint: disable=unused-import
# flake8: noqa

from .base import BaseImageItem, RawImageItem
from .filter import ImageFilterItem, XYImageFilterItem
from .standard import ImageItem, RGBImageItem, XYImageItem
from .masked import MaskedArea, MaskedImageItem, MaskedXYImageItem
from .misc import (
    Histogram2DItem,
    QuadGridItem,
    assemble_imageitems,
    compute_trimageitems_original_size,
    get_image_from_plot,
    get_image_from_qrect,
    get_image_in_shape,
    get_items_in_rectangle,
    get_plot_qrect,
)
from .transform import TrImageItem



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Interactive plotting widgets
============================

Overview
--------

The `pyplot` module provides an interactive plotting interface similar to
`Matplotlib`'s, i.e. with MATLAB-like syntax.

The :py:mod:`.pyplot` module was designed to be as close as possible
to the :py:mod:`matplotlib.pyplot` module, so that one could easily switch
between these two modules by simply changing the import statement. Basically,
if :mod:`plotpy` does support the plotting commands called in your script, replacing
``import matplotlib.pyplot`` by ``import plotpy.pyplot`` should be enough, as
shown in the following example:

* Simple example using `matplotlib`::

    import matplotlib.pyplot as plt
    import numpy as np
    x = np.linspace(-10, 10)
    plt.plot(x, x**2, 'r+')
    plt.show()

* Switching from `matplotlib` to :mod:`plotpy` is trivial::

    import plotpy.pyplot as plt # only this line has changed!
    import numpy as np
    x = np.linspace(-10, 10)
    plt.plot(x, x**2, 'r+')
    plt.show()

Examples
--------

>>> import numpy as np
>>> import plotpy.pyplot as plt
>>> plt.ion() # switching to interactive mode
>>> x = np.linspace(-5, 5, 1000)
>>> plt.figure(1)
>>> plt.subplot(2, 1, 1)
>>> plt.plot(x, np.sin(x), "r+")
>>> plt.plot(x, np.cos(x), "g-")
>>> plt.errorbar(x, -1+x**2/20+.2*np.random.rand(len(x)), x/20)
>>> plt.xlabel("Axe x")
>>> plt.ylabel("Axe y")
>>> plt.subplot(2, 1, 2)
>>> img = np.fromfunction(lambda x, y: np.sin((x/200.)*(y/200.)**2), (1000, 1000))
>>> plt.xlabel("pixels")
>>> plt.ylabel("pixels")
>>> plt.zlabel("intensity")
>>> plt.gray()
>>> plt.imshow(img)
>>> plt.figure("plotyy")
>>> plt.plotyy(x, np.sin(x), x, np.cos(x))
>>> plt.ylabel("sine", "cosine")
>>> plt.show()

Reference
---------

.. autofunction:: interactive
.. autofunction:: ion
.. autofunction:: ioff

.. autofunction:: figure
.. autofunction:: gcf
.. autofunction:: gca
.. autofunction:: show
.. autofunction:: subplot
.. autofunction:: close

.. autofunction:: title
.. autofunction:: xlabel
.. autofunction:: ylabel
.. autofunction:: zlabel

.. autofunction:: yreverse
.. autofunction:: grid
.. autofunction:: legend
.. autofunction:: colormap

.. autofunction:: savefig

.. autofunction:: plot
.. autofunction:: plotyy
.. autofunction:: semilogx
.. autofunction:: semilogy
.. autofunction:: loglog
.. autofunction:: errorbar
.. autofunction:: hist
.. autofunction:: imshow
.. autofunction:: pcolor
"""

# pylint: disable=unused-import
# pylint: disable=wildcard-import
# flake8: noqa

from plotpy.plot.interactive import *



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""Testing `AnnotatedXRange` and `AnnotatedYRange` items"""

# guitest: show

import numpy as np
from guidata.qthelpers import qt_app_context

from plotpy.builder import make
from plotpy.tests import vistools as ptv


def test_annotated_range_selection():
    """Test AnnotatedXRange and AnnotatedYRange items"""
    x = np.linspace(-5, 15, 200)
    y = np.sin(np.sin(np.sin(x))) + 0.5
    with qt_app_context(exec_loop=True):
        items = [
            make.curve(x, y, color="b"),
            make.annotated_xrange(x0=-5, x1=0.0, title="X Range Selection"),
            make.annotated_yrange(y0=-0.2, y1=0.1, title="Y Range Selection"),
        ]
        _win = ptv.show_items(
            items,
            wintitle=test_annotated_range_selection.__doc__,
            title="Annotated Range Selection",
            plot_type="curve",
            disable_readonly_for_items=False,
        )


if __name__ == "__main__":
    test_annotated_range_selection()



# -*- coding: utf-8 -*-

from __future__ import annotations

from typing import TYPE_CHECKING

from guidata.dataset import (
    BoolItem,
    ChoiceItem,
    ColorItem,
    DataSet,
    FloatItem,
    GetAttrProp,
    IntItem,
    StringItem,
)

from plotpy.config import _
from plotpy.styles.base import ItemParameters
from plotpy.styles.image import BaseImageParam

if TYPE_CHECKING:
    from plotpy.items import Histogram2DItem, HistogramItem


class HistogramParam(DataSet):
    n_bins = IntItem(_("Bins"), default=100, min=1, help=_("Number of bins"))
    bin_min = FloatItem(_("Min"), default=None, help=_("Minimum value"), check=False)
    bin_max = FloatItem(_("Max"), default=None, help=_("Maximum value"), check=False)
    logscale = BoolItem(_("logarithmic"), _("Y-axis scale"), default=False)

    def update_param(self, item: HistogramItem) -> None:
        """Update the histogram parameters from the plot item

        Args:
            item: Histogram item
        """
        self.n_bins = item.get_bins()
        self.bin_min, self.bin_max = item.get_bin_range()
        self.logscale = item.get_logscale()

    def update_hist(self, item: HistogramItem) -> None:
        """Update the histogram plot item from the parameters

        Args:
            item: Histogram item
        """
        if self.bin_min is None or self.bin_max is None:
            item.bin_range = None
        else:
            item.bin_range = (self.bin_min, self.bin_max)
        item.bins = self.n_bins
        item.logscale = self.logscale
        item.update_histogram()


class Histogram2DParam(BaseImageParam):
    """Histogram"""

    _multiselection = False
    label = StringItem(_("Title"), default=_("Histogram")).set_prop(
        "display", hide=GetAttrProp("_multiselection")
    )
    nx_bins = IntItem(
        _("X-axis bins"), default=100, min=1, help=_("Number of bins along x-axis")
    )
    ny_bins = IntItem(
        _("Y-axis bins"), default=100, min=1, help=_("Number of bins along y-axis")
    )
    logscale = BoolItem(_("logarithmic"), _("Z-axis scale"), default=False)

    computation = ChoiceItem(
        _("Computation"),
        [
            (-1, _("Bin count")),
            (0, _("Maximum value")),
            (1, _("Mininum value")),
            (2, _("Sum")),
            (3, _("Product")),
            (4, _("Average")),
        ],
        default=-1,
        help=_(
            "Bin count : counts the number of points per bin,\n"
            "For max, min, sum, product, average, compute the "
            "function of a third parameter (one by default)"
        ),
    )
    auto_lut = BoolItem(
        _("Automatic LUT range"),
        default=True,
        help=_("Automatically adapt color scale when panning, zooming"),
    )
    background = ColorItem(
        _("Background color"),
        default="transparent",
        help=_("Background color when no data is present"),
    )

    def update_param(self, item: Histogram2DItem) -> None:
        """Update the histogram parameters from the plot item

        Args:
            item: 2D Histogram item
        """
        super().update_param(item)
        self.logscale = item.logscale
        self.nx_bins, self.ny_bins = item.nx_bins, item.ny_bins

    def update_histogram(self, item: Histogram2DItem) -> None:
        """Update the histogram plot item from the parameters

        Args:
            item: 2D Histogram item
        """
        item.logscale = int(self.logscale)
        item.set_background_color(self.background)
        item.set_bins(self.nx_bins, self.ny_bins)
        self.update_item(item)


class Histogram2DParam_MS(Histogram2DParam):
    _multiselection = True


ItemParameters.register_multiselection(Histogram2DParam, Histogram2DParam_MS)



# -*- coding: utf-8 -*-
#
# Licensed under the terms of the BSD 3-Clause
# (see plotpy/LICENSE for details)

"""
Colormap manager
----------------

This module provides a fully featured colormap manager widget and dialog window
that allows to select, edit and save colormaps (existing or new).


Reference
~~~~~~~~~

.. autoclass:: ColorMapNameEdit
    :members:
.. autoclass:: ColorMapManager
    :members:
"""

from __future__ import annotations

from copy import deepcopy

import qtpy.QtCore as QC
import qtpy.QtGui as QG
import qtpy.QtWidgets as QW
from guidata.configtools import get_icon
from guidata.env import execenv
from guidata.qthelpers import exec_dialog

from plotpy.config import _
from plotpy.mathutils.colormap import (
    ALL_COLORMAPS,
    CUSTOM_COLORMAPS,
    DEFAULT_COLORMAPS,
    LARGE_ICON_HEIGHT,
    LARGE_ICON_ORIENTATION,
    LARGE_ICON_WIDTH,
    add_cmap,
    build_icon_from_cmap,
    cmap_exists,
    delete_cmap,
    get_cmap,
)
from plotpy.widgets.colormap.editor import ColorMapEditor
from plotpy.widgets.colormap.widget import EditableColormap


class ColorMapNameEdit(QW.QDialog):
    """Dialog box to enter a colormap name.

    Args:
        parent: parent QWidget. Defaults to None.
        title: dialog box title. Defaults to "".
        name: default colormap name. Defaults to "".
    """

    def __init__(
        self, parent: QW.QWidget | None = None, title: str = "", name: str = ""
    ) -> None:
        super().__init__(parent)
        self.setWindowTitle(title)

        label = QW.QLabel(_("Enter a colormap name:"))
        self._edit = QW.QLineEdit()
        regexp = QC.QRegularExpression("[1-9a-zA-Z_]*")
        self._edit.setValidator(QG.QRegularExpressionValidator(regexp))
        self._edit.setText(name)
        self._edit.setToolTip(
            _(
                "New colormap name cannot contain special characters "
                "except underscores (_)."
            )
        )

        bbox = QW.QDialogButtonBox(QW.QDialogButtonBox.Ok | QW.QDialogButtonBox.Cancel)
        bbox.accepted.connect(self.accept)
        bbox.rejected.connect(self.reject)
        hlayout = QW.QHBoxLayout()
        hlayout.addWidget(label)
        hlayout.addWidget(self._edit)

        dialog_layout = QW.QVBoxLayout()
        dialog_layout.addLayout(hlayout)
        dialog_layout.addWidget(bbox)
        self.setLayout(dialog_layout)

    def get_colormap_name(self) -> str:
        """Return the colormap name entered in the QLineEdit.

        Returns:
            colormap name
        """
        return self._edit.text()

    @classmethod
    def edit(
        cls, parent: QW.QWidget | None = None, title: str = "", name: str = ""
    ) -> str | None:
        """Open the dialog box and return the colormap name entered in the QLineEdit.

        Args:
            parent: parent QWidget. Defaults to None.
            title: dialog box title. Defaults to "".
            name: default colormap name. Defaults to "".

        Returns:
            colormap name, or None if the dialog box was canceled.
        """
        dialog = cls(parent, title, name)
        if exec_dialog(dialog):
            return dialog.get_colormap_name()
        return None


class ColorMapManager(QW.QDialog):
    """Colormap manager dialog box. Allows to select, edit and save colormaps.

    Args:
        parent: parent QWidget. Defaults to None.
        active_colormap: name of the active colormap.

    .. note::

        The active colormap is the colormap that will be selected by default when the
        dialog box is opened. If the colormap does not exist (or if None is provided),
        the first colormap in the list will be selected by default.

        The active colormap cannot be removed. If the active colormap is a custom
        colormap, the remove button will be enabled but a dialog box will warn the user
        that the colormap cannot be removed.
    """

    SIG_APPLY_COLORMAP = QC.Signal(str)

    def __init__(
        self,
        parent: QW.QWidget | None = None,
        active_colormap: str | None = None,
    ) -> None:
        super().__init__(parent)
        self.setWindowIcon(get_icon("cmap_edit.png"))
        self.setWindowTitle(_("Colormap manager"))

        self.active_cmap_name = default_colormap = active_colormap

        self.__returned_colormap: EditableColormap | None = None

        if default_colormap is None or not cmap_exists(default_colormap, ALL_COLORMAPS):
            default_colormap = next(iter(ALL_COLORMAPS))

        # Select the active colormap
        self._cmap_choice = QW.QComboBox()
        self._cmap_choice.setMaxVisibleItems(15)
        for cmap in ALL_COLORMAPS.values():
            icon = build_icon_from_cmap(
                cmap, LARGE_ICON_WIDTH, LARGE_ICON_HEIGHT, LARGE_ICON_ORIENTATION, 1
            )
            self._cmap_choice.addItem(icon, cmap.name, cmap)

        self._cmap_choice.setIconSize(QC.QSize(LARGE_ICON_WIDTH, LARGE_ICON_HEIGHT))
        self._cmap_choice.setCurrentText(default_colormap)

        add_btn = QW.QPushButton(get_icon("edit_add.png"), _("Add") + "...")
        add_btn.clicked.connect(self.add_colormap)
        self._remove_btn = QW.QPushButton(get_icon("delete.png"), _("Remove") + "...")
        is_custom_cmap = cmap_exists(default_colormap, CUSTOM_COLORMAPS)
        self._remove_btn.setEnabled(is_custom_cmap)
        self._remove_btn.clicked.connect(self.remove_colormap)

        select_gbox = QW.QGroupBox(_("Select or create a colormap"))
        select_label = QW.QLabel(_("Colormap presets:"))
        select_gbox_layout = QW.QHBoxLayout()
        select_gbox_layout.addWidget(select_label)
        select_gbox_layout.addWidget(self._cmap_choice)
        select_gbox_layout.addSpacing(10)
        select_gbox_layout.addWidget(add_btn)
        select_gbox_layout.addWidget(self._remove_btn)
        select_gbox.setLayout(select_gbox_layout)

        # Edit the selected colormap
        current_cmap = self._cmap_choice.currentData()
        # This test is necessary because the currentData method returns a QVariant
        # object, which may be handled differently by PyQt5, PySide2 and PySide6.
        # This is an attempt to fix an issue with PySide6 (segfault).
        if isinstance(current_cmap, EditableColormap):
            current_cmap = deepcopy(current_cmap)
        else:
            current_cmap = None
        self.colormap_editor = ColorMapEditor(self, colormap=current_cmap)

        self._edit_gbox = QW.QGroupBox(_("Edit the selected colormap"))
        edit_gbox_layout = QW.QVBoxLayout()
        edit_gbox_layout.setContentsMargins(0, 0, 0, 0)
        edit_gbox_layout.addWidget(self.colormap_editor)
        self._edit_gbox.setLayout(edit_gbox_layout)
        self._edit_gbox.setCheckable(True)
        self._edit_gbox.setChecked(False)
        self.colormap_editor.colormap_widget.COLORMAP_CHANGED.connect(
            self._changes_not_saved
        )

        self._cmap_choice.currentIndexChanged.connect(self.set_colormap)

        self.bbox = QW.QDialogButtonBox(
            QW.QDialogButtonBox.Apply
            | QW.QDialogButtonBox.Save
            | QW.QDialogButtonBox.Ok
            | QW.QDialogButtonBox.Cancel
        )
        self._changes_saved = True
        self._save_btn = self.bbox.button(QW.QDialogButtonBox.Save)
        self._save_btn.setEnabled(False)  # type: ignore
        self._apply_btn = self.bbox.button(QW.QDialogButtonBox.Apply)
        self._apply_btn.setEnabled(False)  # type: ignore
        self.bbox.clicked.connect(self.button_clicked)

        dialog_layout = QW.QVBoxLayout()
        dialog_layout.addWidget(select_gbox)
        dialog_layout.addWidget(self._edit_gbox)
        dialog_layout.addWidget(self.bbox)
        self.setLayout(dialog_layout)

        # The dialog needs to be resizable horizontally but not vertically:
        self.setFixedHeight(self.sizeHint().height())

    def button_clicked(self, button: QW.QAbstractButton) -> None:
        """Callback function to be called when a button is clicked.

        Args:
            button: button clicked
        """
        if button is self._apply_btn:
            if not self.current_changes_saved and not self.save_colormap():
                return
            self._apply_btn.setEnabled(False)
            self.SIG_APPLY_COLORMAP.emit(self.colormap_editor.get_colormap().name)
        elif button is self._save_btn:
            self.save_colormap()
        elif self.bbox.buttonRole(button) == QW.QDialogButtonBox.AcceptRole:
            self.accept()
        else:
            self.reject()

    def _changes_not_saved(self) -> None:
        """Callback function to be called when the colormap is modified. Enables the
        save button and sets the current_changes_saved attribute to False."""
        self._save_btn.setEnabled(True)  # type: ignore
        self._apply_btn.setEnabled(True)  # type: ignore
        self._changes_saved = False

    @property
    def current_changes_saved(self) -> bool:
        """Getter to know if the current colormap has been saved or not.

        Returns:
            True if the current colormap has been saved, False otherwise.
        """
        return self._changes_saved

    def set_colormap(self, index: int) -> None:
        """Set the current colormap to the value present at the given index in the
        QComboBox. Makes a copy of the colormap object so the ColorMapEditor does not
        mutate the original colormap object.

        Args:
            index: index of the colormap in the QComboBox.
        """
        cmap_copy: EditableColormap = deepcopy(self._cmap_choice.itemData(index))
        self.colormap_editor.set_colormap(cmap_copy)

        is_custom_cmap = cmap_exists(cmap_copy.name, CUSTOM_COLORMAPS)
        self._remove_btn.setEnabled(is_custom_cmap)

        self._changes_saved = True
        is_new_colormap = not cmap_exists(cmap_copy.name)
        self._save_btn.setEnabled(is_new_colormap)  # type: ignore

    def get_colormap(self) -> EditableColormap | None:
        """Return the selected colormap object.

        Returns:
            selected colormap object
        """
        return self.__returned_colormap

    def __get_new_colormap_name(self, title: str, name: str) -> str | None:
        """Return a new colormap name that does not already exists in the list of
        colormaps.

        Args:
            title: title of the current action
            name: colormap name to check

        Returns:
            new colormap name, or None if the user canceled the action.
        """
        new_name = name
        while True:
            new_name = ColorMapNameEdit.edit(self, title, new_name)
            if new_name is None:
                return None
            if cmap_exists(new_name, DEFAULT_COLORMAPS):
                QW.QMessageBox.warning(
                    self,
                    title,
                    _(
                        "Name <b>%s</b> is already used by a predefined colormap, and "
                        "cannot be used for a custom colormap.<br><br>"
                        "Please choose another name."
                    )
                    % new_name,
                )
                continue
            if cmap_exists(new_name, CUSTOM_COLORMAPS) and (
                QW.QMessageBox.question(
                    self,
                    title,
                    _(
                        "Name <b>%s</b> is already used by a custom colormap.<br><br>"
                        "Do you want to overwrite it?"
                    )
                    % new_name,
                    QW.QMessageBox.Yes | QW.QMessageBox.No,
                    QW.QMessageBox.No,
                )
                == QW.QMessageBox.No
            ):
                continue
            break
        return new_name

    def add_colormap(self) -> None:
        """Create a new colormap and set it as the current colormap."""
        cmap = EditableColormap(QG.QColor(0), QG.QColor(4294967295), name=_("New"))
        if self.save_colormap(cmap):
            # If the colormap save dialog was accepted, then we enable the colormap
            # editor group box.
            # Otherwise, we leave the colormap editor group box at its current state.
            self._edit_gbox.setChecked(True)

    def remove_colormap(self) -> None:
        """Remove the current colormap."""
        cmap = self.colormap_editor.get_colormap()
        if cmap is None:
            return
        name = cmap.name
        if name == self.active_cmap_name or cmap_exists(name, DEFAULT_COLORMAPS):
            if execenv.unattended:  # For testing purposes only
                return
            if name == self.active_cmap_name:
                msg = _("Colormap <b>%s</b> is the active colormap.")
            else:
                msg = _("Colormap <b>%s</b> is a predefined colormap.")
            QW.QMessageBox.warning(
                self,
                _("Remove"),
                msg % name + "<br>" + _("Thus, this colormap cannot be removed."),
                QW.QMessageBox.Ok,
            )
            return
        if execenv.unattended:  # For testing purposes only
            if not execenv.accept_dialogs:
                return
        elif (
            QW.QMessageBox.question(
                self,
                _("Remove"),
                _("Do you want to delete colormap <b>%s</b>?") % name,
                QW.QMessageBox.Yes | QW.QMessageBox.No,
                QW.QMessageBox.No,
            )
            == QW.QMessageBox.No
        ):
            return
        delete_cmap(cmap)
        current_index = self._cmap_choice.currentIndex()
        self._cmap_choice.removeItem(current_index)
        self._changes_saved = True
        self._save_btn.setEnabled(False)
        self._cmap_choice.setCurrentIndex(max(current_index - 1, 0))

    def save_colormap(self, cmap: EditableColormap | None = None) -> bool:
        """Saves the current colormap and handles the validation process. The saved
        colormaps can only be saved in the custom colormaps.

        Args:
            cmap: colormap to save. If None, will use the current colormap.

        Returns:
            True if the colormap has been saved, False otherwise.
        """
        if cmap is None:
            cmap = self.colormap_editor.get_colormap()
            title = _("Save colormap")
        else:
            title = _("Add colormap")
        new_name = self.__get_new_colormap_name(title, cmap.name)
        if new_name is None:
            return False

        # Before modifying CUSTOM_COLORMAPS, save this boolean expression:
        is_existing_custom_cmap = cmap_exists(new_name, CUSTOM_COLORMAPS)

        # Take into account the case where the color map exists in custom colormaps
        # but not with the same case (e.g. "viridis" vs "Viridis"). So we need to
        # get the original name of the colormap:
        if is_existing_custom_cmap:
            new_name = get_cmap(new_name).name

        cmap.name = new_name
        add_cmap(cmap)

        icon = build_icon_from_cmap(
            cmap, LARGE_ICON_WIDTH, LARGE_ICON_HEIGHT, LARGE_ICON_ORIENTATION, 1
        )
        if is_existing_custom_cmap:
            self._cmap_choice.setCurrentText(new_name)
            current_index = self._cmap_choice.currentIndex()
            self._cmap_choice.setItemData(current_index, cmap)
            self._cmap_choice.setItemIcon(current_index, icon)
        else:
            self._cmap_choice.addItem(icon, new_name, cmap)
            self._cmap_choice.setCurrentText(new_name)

        self._save_btn.setEnabled(False)  # type: ignore
        self._changes_saved = True
        return True

    def accept(self) -> None:
        """Adds logic on top of the normal QDialog.accept method to handle colormap
        save.
        """
        if not self.current_changes_saved and not self.save_colormap():
            return
        self.__returned_colormap = self.colormap_editor.get_colormap()
        super().accept()
